#!/usr/bin/python3.7
# type: ignore
########################################################################################
# analysis.py - The analysis component for the model.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################
"""
Used for analysis of the output of the model runs.

NOTE: The mypy type checker is instructed to ignore this component. This is done due to
the lower standards applied to the analysis code, and the failure of mypy to correctly
type-check the external matplotlib.pyplot module.

"""

import argparse
import collections
import os
import sys

from logging import Logger
from typing import Any, List, Dict, Optional, Tuple

import json
import re

import numpy

from matplotlib import pyplot as plt

try:
    from ..__utils__ import get_logger
    from ..pvt_system.constants import (  # pylint: disable=unused-import
        HEAT_CAPACITY_OF_WATER,
    )
    from .__utils__ import (
        GraphDetail,
        load_model_data,
        plot_figure,
    )
except ModuleNotFoundError:
    import logging

    logging.error(
        "Incorrect module import. Try running with `python3.7 -m pvt_model.analysis`"
    )
    raise

__all__ = ("analyse",)

# Used to distinguish copled data sets.
COUPLED_DATA_TYPE = "coupled"
# Used to distinguished decoupled data sets.
DECOUPLED_DATA_TYPE = "decoupled"
# Used to distinguish dynamic data sets.
DYNAMIC_DATA_TYPE = "dynamic"
# The directory into which which should be saved
NEW_FIGURES_DIRECTORY: str = "figures"
# Used to identify the "time" data set.
TIME_KEY = "time"
# The directory in which old figures are saved and stored for long-term access
OLD_FIGURES_DIRECTORY: str = "old_figures"
# Used to specify the reduced temperature for comparing the thermal performance of the
# collectors.
REDUCED_TEMPERATURE_COMPARISON_POINT = 0.10
# Used to distinguish steady-state data sets.
STEADY_STATE_DATA_TYPE = "steady_state"
# Name of the steady-state data file.
STEADY_STATE_DATA_FILE_NAME = "autotherm.yaml"
# How detailed the graph should be
GRAPH_DETAIL: GraphDetail = GraphDetail.lowest
# How many values there should be between each tick on the x-axis
# X_TICK_SEPARATION: int = int(8 * GRAPH_DETAIL.value / 48)
X_TICK_SEPARATION: int = 8
# Which days of data to include
DAYS_TO_INCLUDE: List[bool] = [False, True]
# Portion-covered regex.
MASS_FLOW_RATE_REGEX = re.compile(
    r"(?P<panel_type>[^_]*)_(?P<no_pv>no_pv_|)"
    r"(?P<glazing>single_|double_|un)glazed_(?P<pv>[^\d]*|)"
    r"(?P<mass_flow_rate>\d*_\d)_litres_per_hour_(?P<x_resolution>\d*)_x_"
    r"(?P<y_resolution>\d*).*"
)
# A regex is needed to exclude results with no pv layer.
NO_PV_REGEX = re.compile(r".*no_pv.*")


def _calculate_value_at_zero_reduced_temperature(
    filedata: Dict[Any, Any], parameter: str
) -> float:
    """
    Calculates the value of the parameter passed in at a reduced temperature of zero.

    :param filedata:
        The raw data extracted from the data file.

    :param parameter:
        A `str` specifying the parameter that should be computed at a reduced
        temperature of zero.

    :return:
        The value of the parameter specified at a reduced temperature of zero.

    """

    x_series: List[float] = []
    y_series: List[float] = []
    for key, value in filedata.items():
        x_series.append(float(key))
        y_series.append(float(value[parameter]))

    trend = numpy.polyfit(x_series, y_series, 2)
    return (
        trend[2] * REDUCED_TEMPERATURE_COMPARISON_POINT ** 2
        + trend[1] * REDUCED_TEMPERATURE_COMPARISON_POINT
        + trend[2]
    )


def _calculate_zero_point_efficiencies(filedata: Dict[Any, Any]) -> Tuple[float, float]:
    """
    Calculates the various values at a reduced temperature of zero Kelvin.

    :param filedata:
        The raw data extracted from the data file.

    :return:
        A `tuple` containing:
        - the electrical efficiency,
        - and the thermal efficiency
        at a reduced temperature of zero.

    """

    return (
        _calculate_value_at_zero_reduced_temperature(filedata, "electrical_efficiency"),
        _calculate_value_at_zero_reduced_temperature(filedata, "thermal_efficiency"),
    )


def _parse_args(args) -> argparse.Namespace:
    """
    Parse the CLI args.

    """

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--data-file-directory",
        "-dfdir",
        help="Path to the directory containing data files to parse.",
    )
    parser.add_argument(
        "--show-output",
        "-so",
        action="store_true",
        default=False,
        help="Show the output figures generated.",
    )

    return parser.parse_args(args)


def _resolution_from_graph_detail(
    graph_detail: GraphDetail, num_data_points: int
) -> int:
    """
    Determine the x-step resolution of the plot.

    I realise here that the word "resolution" is over-used. Here, the number of data
    points (in the system data) that need to be absorbed into one graph point is
    calculated and returned.

    :param graph_detail:
        The detail level needed on the graph.

    :param num_data_points:
        The number of data points recorded in the plot.

    :return:
        The number of data points that need to be absorbed per graph point.

    """

    # * For "lowest", include one point every half-hour.
    if graph_detail == GraphDetail.highest:
        return 1

    return int(num_data_points / graph_detail.value)


def _post_process_data(
    data_to_post_process: Dict[str, Dict[Any, Any]]
) -> Dict[str, Dict[Any, Any]]:
    """
    Carries out post-processing on data where necessary.

    Things to be computed:
        - Bulk Water Temperautre
        - Litres consumed

    :param data_to_post_process:
        The data to be post processed.

    :return:
        The post-processed data.

    """

    # * Cycle through all the data points and compute the new values as needed.
    for data_entry in data_to_post_process.values():
        data_entry["absorber_temperature_gain"] = (
            data_entry["collector_output_temperature"]
            - data_entry["collector_input_temperature"]
        )
    return data_to_post_process


def analyse_decoupled_steady_state_data(  # pylint: disable=too-many-branches
    data: Dict[Any, Dict[str, Any]], logger: Logger
) -> None:
    """
    Carry out analysis on a series of of steady-state data sets.

    :param data:
        The data to analyse. This should be a `dict` mapping the file name to the data
        extracted from the file.

    :param logger:
        The logger to use for the analysis run.

    """

    logger.info("Beginning steady-state analysis.")

    # Construct a reduced mapping
    autotherm_single_glazed_data: Dict[str, Any] = collections.defaultdict(dict)
    autotherm_no_pv_single_glazed_data: Dict[str, Any] = collections.defaultdict(dict)
    autotherm_double_glazed_data: Dict[str, Any] = collections.defaultdict(dict)
    autotherm_no_pv_double_glazed_data: Dict[str, Any] = collections.defaultdict(dict)
    ilaria_single_glazed_data: Dict[str, Any] = collections.defaultdict(dict)
    ilaria_no_pv_single_glazed_data: Dict[str, Any] = collections.defaultdict(dict)
    ilaria_double_glazed_data: Dict[str, Any] = collections.defaultdict(dict)
    ilaria_no_pv_double_glazed_data: Dict[str, Any] = collections.defaultdict(dict)
    for key, sub_dict in data.items():
        mass_flow_rate_match = re.match(MASS_FLOW_RATE_REGEX, key)
        if mass_flow_rate_match is None:
            continue

        mass_flow_rate_string: str = mass_flow_rate_match.group("mass_flow_rate")
        try:
            mass_flow_rate: float = int(
                mass_flow_rate_string.split("_")[0]
            ) + 0.1 * int(mass_flow_rate_string.split("_")[1])
        except ValueError:
            mass_flow_rate = int(mass_flow_rate_string.split("_")[1])

        electrical_efficiency, thermal_efficiency = _calculate_zero_point_efficiencies(
            sub_dict
        )

        # If there was no pv layer present in the simulation.
        if mass_flow_rate_match.group("no_pv") == "":
            if mass_flow_rate_match.group("panel_type") == "autotherm":
                if f"{mass_flow_rate_match.group('glazing')}glazed" == "single_glazed":
                    autotherm_single_glazed_data[mass_flow_rate][
                        "electrical_efficiency"
                    ] = electrical_efficiency
                    autotherm_single_glazed_data[mass_flow_rate][
                        "thermal_efficiency"
                    ] = thermal_efficiency
                elif (
                    f"{mass_flow_rate_match.group('glazing')}glazed" == "double_glazed"
                ):
                    autotherm_double_glazed_data[mass_flow_rate][
                        "electrical_efficiency"
                    ] = electrical_efficiency
                    autotherm_double_glazed_data[mass_flow_rate][
                        "thermal_efficiency"
                    ] = thermal_efficiency

            if mass_flow_rate_match.group("panel_type") == "ilaria":
                if f"{mass_flow_rate_match.group('glazing')}glazed" == "single_glazed":
                    ilaria_single_glazed_data[mass_flow_rate][
                        "electrical_efficiency"
                    ] = electrical_efficiency
                    ilaria_single_glazed_data[mass_flow_rate][
                        "thermal_efficiency"
                    ] = thermal_efficiency
                elif (
                    f"{mass_flow_rate_match.group('glazing')}glazed" == "double_glazed"
                ):
                    ilaria_double_glazed_data[mass_flow_rate][
                        "electrical_efficiency"
                    ] = electrical_efficiency
                    ilaria_double_glazed_data[mass_flow_rate][
                        "thermal_efficiency"
                    ] = thermal_efficiency
        else:
            if mass_flow_rate_match.group("panel_type") == "autotherm":
                if f"{mass_flow_rate_match.group('glazing')}glazed" == "single_glazed":
                    autotherm_no_pv_single_glazed_data[mass_flow_rate][
                        "electrical_efficiency"
                    ] = electrical_efficiency
                    autotherm_no_pv_single_glazed_data[mass_flow_rate][
                        "thermal_efficiency"
                    ] = thermal_efficiency
                elif (
                    f"{mass_flow_rate_match.group('glazing')}glazed" == "double_glazed"
                ):
                    autotherm_no_pv_double_glazed_data[mass_flow_rate][
                        "electrical_efficiency"
                    ] = electrical_efficiency
                    autotherm_no_pv_double_glazed_data[mass_flow_rate][
                        "thermal_efficiency"
                    ] = thermal_efficiency

            if mass_flow_rate_match.group("panel_type") == "ilaria":
                if f"{mass_flow_rate_match.group('glazing')}glazed" == "single_glazed":
                    ilaria_no_pv_single_glazed_data[mass_flow_rate][
                        "electrical_efficiency"
                    ] = electrical_efficiency
                    ilaria_no_pv_single_glazed_data[mass_flow_rate][
                        "thermal_efficiency"
                    ] = thermal_efficiency
                elif (
                    f"{mass_flow_rate_match.group('glazing')}glazed" == "double_glazed"
                ):
                    ilaria_no_pv_double_glazed_data[mass_flow_rate][
                        "electrical_efficiency"
                    ] = electrical_efficiency
                    ilaria_no_pv_double_glazed_data[mass_flow_rate][
                        "thermal_efficiency"
                    ] = thermal_efficiency

    # Plot the PV thermal efficiencies.
    plot_figure(
        "autotherm_single_glazed_thermal_efficiency_against_mass_flow_rate",
        autotherm_single_glazed_data,
        ["thermal_efficiency"],
        "Thermal efficiency",
        x_axis_label="Mass-flow rate / Litres per hour",
        use_data_keys_as_x_axis=True,
        disable_lines=True,
    )

    plot_figure(
        "autotherm_double_glazed_thermal_efficiency_against_mass_flow_rate",
        autotherm_double_glazed_data,
        ["thermal_efficiency"],
        "Thermal efficiency",
        x_axis_label="Mass-flow rate / Litres per hour",
        use_data_keys_as_x_axis=True,
        disable_lines=True,
    )

    plot_figure(
        "ilaria_single_glazed_thermal_efficiency_against_mass_flow_rate",
        ilaria_single_glazed_data,
        ["thermal_efficiency"],
        "Thermal efficiency",
        x_axis_label="Mass-flow rate / Litres per hour",
        use_data_keys_as_x_axis=True,
        disable_lines=True,
    )

    plot_figure(
        "ilaria_double_glazed_thermal_efficiency_against_mass_flow_rate",
        ilaria_double_glazed_data,
        ["thermal_efficiency"],
        "Thermal efficiency",
        x_axis_label="Mass-flow rate / Litres per hour",
        use_data_keys_as_x_axis=True,
        disable_lines=True,
    )

    # Plot the no-PV thermal efficiencies.
    plot_figure(
        "autotherm_no_pv_single_glazed_thermal_efficiency_against_mass_flow_rate",
        autotherm_no_pv_single_glazed_data,
        ["thermal_efficiency"],
        "Thermal efficiency",
        x_axis_label="Mass-flow rate / Litres per hour",
        use_data_keys_as_x_axis=True,
        disable_lines=True,
    )

    plot_figure(
        "autotherm_no_pv_double_glazed_thermal_efficiency_against_mass_flow_rate",
        autotherm_no_pv_double_glazed_data,
        ["thermal_efficiency"],
        "Thermal efficiency",
        x_axis_label="Mass-flow rate / Litres per hour",
        use_data_keys_as_x_axis=True,
        disable_lines=True,
    )

    plot_figure(
        "ilaria_no_pv_single_glazed_thermal_efficiency_against_mass_flow_rate",
        ilaria_no_pv_single_glazed_data,
        ["thermal_efficiency"],
        "Thermal efficiency",
        x_axis_label="Mass-flow rate / Litres per hour",
        use_data_keys_as_x_axis=True,
        disable_lines=True,
    )

    plot_figure(
        "ilaria_no_pv_double_glazed_thermal_efficiency_against_mass_flow_rate",
        ilaria_no_pv_double_glazed_data,
        ["thermal_efficiency"],
        "Thermal efficiency",
        x_axis_label="Mass-flow rate / Litres per hour",
        use_data_keys_as_x_axis=True,
        disable_lines=True,
    )

    # Plot the pv electrical efficiencies.
    plot_figure(
        "autotherm_single_glazed_electrical_efficiency_against_mass_flow_rate",
        autotherm_single_glazed_data,
        ["electrical_efficiency"],
        "Electrical efficiency",
        x_axis_label="Mass-flow rate / Litres per hour",
        use_data_keys_as_x_axis=True,
        disable_lines=True,
    )

    plot_figure(
        "autotherm_double_glazed_electrical_efficiency_against_mass_flow_rate",
        autotherm_double_glazed_data,
        ["electrical_efficiency"],
        "Electrical efficiency",
        x_axis_label="Mass-flow rate / Litres per hour",
        use_data_keys_as_x_axis=True,
        disable_lines=True,
    )

    plot_figure(
        "ilaria_single_glazed_electrical_efficiency_against_mass_flow_rate",
        ilaria_single_glazed_data,
        ["electrical_efficiency"],
        "Electrical efficiency",
        x_axis_label="Mass-flow rate / Litres per hour",
        use_data_keys_as_x_axis=True,
        disable_lines=True,
    )

    plot_figure(
        "ilaria_double_glazed_electrical_efficiency_against_mass_flow_rate",
        ilaria_double_glazed_data,
        ["electrical_efficiency"],
        "Electrical efficiency",
        x_axis_label="Mass-flow rate / Litres per hour",
        use_data_keys_as_x_axis=True,
        disable_lines=True,
    )

    # Plot the no-PV electrical efficiencies.
    plot_figure(
        "autotherm_no_pv_single_glazed_electrical_efficiency_against_mass_flow_rate",
        autotherm_no_pv_single_glazed_data,
        ["electrical_efficiency"],
        "Electrical efficiency",
        x_axis_label="Mass-flow rate / Litres per hour",
        use_data_keys_as_x_axis=True,
        disable_lines=True,
    )

    plot_figure(
        "autotherm_no_pv_double_glazed_electrical_efficiency_against_mass_flow_rate",
        autotherm_no_pv_double_glazed_data,
        ["electrical_efficiency"],
        "Electrical efficiency",
        x_axis_label="Mass-flow rate / Litres per hour",
        use_data_keys_as_x_axis=True,
        disable_lines=True,
    )

    plot_figure(
        "ilaria_no_pv_single_glazed_electrical_efficiency_against_mass_flow_rate",
        ilaria_no_pv_single_glazed_data,
        ["electrical_efficiency"],
        "Electrical efficiency",
        x_axis_label="Mass-flow rate / Litres per hour",
        use_data_keys_as_x_axis=True,
        disable_lines=True,
    )

    plot_figure(
        "ilaria_no_pv_double_glazed_electrical_efficiency_against_mass_flow_rate",
        ilaria_no_pv_double_glazed_data,
        ["electrical_efficiency"],
        "Electrical efficiency",
        x_axis_label="Mass-flow rate / Litres per hour",
        use_data_keys_as_x_axis=True,
        disable_lines=True,
    )


def analyse(data_file_directory: str, show_output: Optional[bool] = False) -> None:
    """
    The main method for the analysis module.

    :param data_file_directory:
        The path to the directory containing data files to analyse.

    :param show_output:
        Whether to show the output files generated.

    """

    # * Set up the logger
    logger = get_logger(True, "pvt_analysis", True)

    # * Extract the data.
    data: Dict[str, Dict[str, Any]] = dict()
    for data_file_name in os.listdir(data_file_directory):
        try:
            data[data_file_name] = load_model_data(
                os.path.join(data_file_directory, data_file_name)
            )
        except json.JSONDecodeError:
            print(f"JSON decode error encountered. File name: {data_file_name}")
            raise

        # If the data type is not decoupled, steady-state data, then exit.
        data_type = data[data_file_name].pop("data_type")
        if data_type not in (
            STEADY_STATE_DATA_TYPE,
            f"{DECOUPLED_DATA_TYPE}_{STEADY_STATE_DATA_TYPE}",
        ):
            logger.info(
                "Data type was neither 'dynamic' nor 'steady_state'. Omitting..."
            )
            data.pop(data_file_name)
            continue

    analyse_decoupled_steady_state_data(data, logger)

    logger.info("Analysis complete - all figures saved successfully.")

    if show_output:
        plt.show()


if __name__ == "__main__":
    parsed_args = _parse_args(sys.argv[1:])
    analyse(parsed_args.data_file_directory, parsed_args.show_output)



#!/usr/bin/python3.7
########################################################################################
# __utils__.py - The utility module for this, my first, PV-T model! :O
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2020
########################################################################################

"""
The utility module for the PV-T model.

This module contains common functionality, strucutres, and types, to be used by the
various modules throughout the PVT model.

"""

import datetime
import enum

from typing import Any, Dict, Generator, List, Union

from dataclasses import dataclass
from dateutil.relativedelta import relativedelta

import numpy

from .constants import SPECIFIC_GAS_CONSTANT_OF_AIR


from ..__utils__ import (
    ProgrammerJudgementFault,
)

__all__ = (
    "BaseDailyProfile",
    "CollectorParameters",
    "Date",
    "DivergentSolutionError",
    "InternalError",
    "InvalidDataError",
    "MissingDataError",
    "ResolutionMismatchError",
    "LayerParameters",
    "OpticalLayerParameters",
    "PVParameters",
    "PVT_SYSTEM_MODEL_LOGGER_NAME",
    "TEMPERATURE_FOLDERNAME",
    "time_iterator",
    "UtilityType",
    "WeatherConditions",
    "WEATHER_DATA_FILENAME",
)


#############
# Constants #
#############

# Name of the logger used
PVT_SYSTEM_MODEL_LOGGER_NAME = "pvt_system.{tag}_run_{run_number}"
# Folder containing the solar irradiance profiles
SOLAR_IRRADIANCE_FOLDERNAME = "solar_irradiance_profiles"
# Folder containing the temperature profiles
TEMPERATURE_FOLDERNAME = "temperature_profiles"
# Name of the weather data file.
WEATHER_DATA_FILENAME = "weather.yaml"

##############
# Exceptions #
##############


class DivergentSolutionError(Exception):
    """
    Raised when a divergent solution occurs.

    """

    def __init__(
        self,
        convergence_run_number: int,
        run_one_temperature_difference: float,
        run_one_temperature_vector: Union[List[float], numpy.ndarray],
        run_two_temperature_difference: float,
        run_two_temperature_vector: Union[List[float], numpy.ndarray],
    ) -> None:
        """
        Instantiate a :class:`DivergentSolutionError`.

        :param convergence_run_number:
            The number of runs attempted to reach a convergent solution.

        :param run_one_temperature_difference:
            The temperature difference between the i-2 and i-1 iterations.

        :param run_one_temperature_vector:
            The temperature vector computed at the i-1 iteration.

        :param run_two_temperature_difference:
            The temperature difference between the i-1 and i iterations.

        :param run_two_temperature_vector:
            The temperature vector computed at the i iteration.

        """

        run_one_temperature_vector_string = ""
        for entry in run_one_temperature_vector:
            run_one_temperature_vector_string += ", {}".format(str(entry))

        run_two_temperature_vector_string = ""
        for entry in run_two_temperature_vector:
            run_two_temperature_vector_string += ", {}".format(str(entry))

        super().__init__(
            "A divergent solution was found when attempting to compute the "
            "temperatures at the next time step:\n"
            f"Number of convergent runs attempted: {convergence_run_number}\n"
            f"Previous difference: {run_one_temperature_difference}\n"
            f"Current difference: {run_two_temperature_difference}\n"
            "Divergence is hence "
            f"{round(run_two_temperature_difference - run_one_temperature_difference, 2)}"
            " away from the current solution.\n"
            f"Previous solution temperatures:\n{run_one_temperature_vector_string}\n"
            f"Current solution temperatures:\n{run_two_temperature_vector_string}\n",
        )


class InternalError(Exception):
    """
    Used for internal error handling and catching where no information is needed.

    """


class InvalidDataError(Exception):
    """
    Raised when some proviced data is of the wrong format.

    """

    def __init__(self, data_file_name: str, message: str) -> None:
        """
        Instantiate an Invalid Data Error.

        :param data_file_name:
            The name of the data file containing the invalid data entry.

        :param message:
            An error message to be appended.

        """

        super().__init__(
            "Invalid data error. "
            "The file '{}' contained data of the wrong format: {}".format(
                data_file_name, message
            )
        )


class MissingDataError(Exception):
    """
    Raised when data is requested from a class that is missing.

    """

    def __init__(self, message: str) -> None:
        """
        Instantiate a Missing Data Error.

        :param message:
            The message to append for the user.

        """

        super().__init__(
            "Data requested from a class could not be found: {}".format(message)
        )


class ResolutionMismatchError(Exception):
    """
    Raised when an error occurs attempting to match up data and simulation resolutions.

    """

    def __init__(self, message: str) -> None:
        """
        Instantiate a resolution mismatch error.

        :param message:
            An appended message to display to the user.

        """

        super().__init__(
            "Due to the nature of the load data, an integer multiple of resolutions, "
            "or divsions of resolutions, must be supplied with the '--resolution' or "
            "'-r' flag.\nI appreciate that this is poor coding, but at least I took "
            "the time to write a custom exception for it :p .\n Error message: "
            f"{message}"
        )


##############################
# Functions and Data Classes #
##############################


class BaseDailyProfile:
    """
    Represents a day's profile.

    """

    # Private Attributes:
    # .. attribute:: _profile
    #   A mapping of the time of day to a floating point value.
    #

    def __init__(self, profile: Dict[datetime.time, float] = None) -> None:
        """
        Instantiate the daily profile class..

        :param profile:
            The daily profile as a mapping between the time of day and the floating
            point value.

        """

        if profile is None:
            profile = dict()

        if not isinstance(profile, dict):
            raise ProgrammerJudgementFault(
                "The input daily profile provided is not a mapping of the correct type."
            )

        self._profile = profile

    def __getitem__(self, index: datetime.time) -> float:
        """
        Return an irradiance value from the profile.

        :param index:
            The index of the item to return, must be a valid :class:`datetime.time`
            instance.

        :return:
            The profile's value at that time.

        """

        # If the index is in the profile, return the index.
        if index in self._profile:
            return self._profile[index]

        # If the index is not in the profile, then the closest value needs to be
        # determined. If there is a tie, this does not matter.
        delta_t_to_t_map = {
            (
                abs(
                    time.hour * 3600
                    + time.minute * 60
                    + time.second
                    - (index.hour * 3600 + index.minute * 60 + index.second)
                )
            ): time
            for time in self._profile
        }
        return self._profile[delta_t_to_t_map[min(delta_t_to_t_map)]]

    def __setitem__(self, index: datetime.time, value: float) -> None:
        """
        Sets an item in the profile.

        :param index:
            The index, i.e., :class:`datetime.time` instance, for which to set the
            value.

        :param value:
            The value to set.

        """

        self._profile[index] = value

    def update(self, profile: Dict[datetime.time, float]) -> None:
        """
        Updates the internal profile with the mapping provided.

        :param profile:
            The profile to add to the currently-stored internal profile.

        """

        if self._profile is None:
            self._profile = profile
        else:
            self._profile.update(profile)


@dataclass
class Date:
    """
    Represents a date, containing informaiton about the month and day.

    .. attribute:: day
        The day of the month, ranging from 1 to 31, expressed as an `int`.

    .. attribute:: month
        The month of the year, expressed as an `int`.

    """

    day: int
    month: int

    def __hash__(self) -> int:
        """
        Returns a hash of the :class:`Date`.

        This is calulated using the inbuilt hashability of the datetime.date module.

        :return:
            A unique hash.

        """

        return hash(datetime.date(2000, self.month, self.day))

    @classmethod
    def from_date(cls, date: datetime.date) -> Any:
        """
        Instantiates a :class:`Date` instance based on a :class:`datetime.date` object.

        :param date:
            The date from which to instantiate this :class:`Date` instance.

        :return:
            An instantiated :class:`Date` instanced based on the information attached
            to the :class:`datetime.date` instance passed in.

        """

        return cls(date.day, date.month)


class UtilityType(enum.Enum):
    """
    Contains information about the type of mains utility being used.

    """

    gas = 0
    electricity = 1


@dataclass
class WeatherConditions:
    """
    Contains information about the various weather conditions at any given time.

    .. attribute:: ambient_tank_temperature
        The ambient temperature surrounding the hot-water tank, measured in Kelvin.

    .. attribute:: ambient_temperature
        The ambient temperature in

    .. attribute:: declination
        The angle of declination of the sun above the horizon

    .. attribute:: azimuthal_angle
        The azimuthal angle of the sun, defined clockwise from True North.

    .. attribute:: mains_water_temperature
        The temperature of the mains water, measured in Kelvin.

    .. attribute:: pressure
        The air pressure, measured in Pascals = Newtons per meter squared.

    .. attribute:: wind_speed
        The wind speed in meters per second.


    """

    # Private attributes:
    #
    # .. attribute:: _irradiance
    #   The solar irradiance in Watts per meter squared.
    #

    _irradiance: float
    ambient_tank_temperature: float
    ambient_temperature: float
    azimuthal_angle: float
    declination: float
    mains_water_temperature: float
    pressure: float
    wind_speed: float

    @property
    def density_of_air(self) -> float:
        """
        The density of air varies as a function of temperature.

        The data for the density is obtained from:
        https://onlinelibrary.wiley.com/doi/pdf/10.1002/9780470516430.app2

        :return:
            The density of air, measured in kilograms per meter cubed.

        """

        return self.pressure / (SPECIFIC_GAS_CONSTANT_OF_AIR * self.ambient_temperature)

    @property
    def dynamic_viscosity_of_air(self) -> float:
        """
        The dynamic viscosity of air varies as a function of temperature.

        The data for the dynamic viscosity is obtained from:
        https://onlinelibrary.wiley.com/doi/pdf/10.1002/9780470516430.app2

        :return:
            The dynamic viscosity of air, measured in kilograms per meter second.

        """

        return (1.458 * (10 ** (-6)) * (self.ambient_temperature ** 1.5)) / (
            self.ambient_temperature + 110.4
        )

    @property
    def heat_capacity_of_air(self) -> float:
        """
        Return the heat capacity of air in Joules perkilogram Kelvin.

        The heat capacity of air varies with a function of temperature and is given by
        an empirically-derived formula.

        https://onlinelibrary.wiley.com/doi/pdf/10.1002/9780470516430.app2

        :return:
            The heat capacity of air in Joules per kilogram Kelvin.

        """

        return 1002.5 + 275 * (10 ** (-6)) * (self.ambient_temperature - 200) ** 2

    @property
    def irradiance(self) -> float:
        """
        The irradiance should only be definied if the sun is above the horizon.

        :return:
            The solar irradiance, adjusted for the day-night cycle, measured in Watts
            per meter squared.

        """

        if self.declination > 0:
            return self._irradiance
        return 0

    @property
    def kinematic_viscosity_of_air(self) -> float:
        """
        The kinematic viscosity of air varies as a function of temperature.

        The data for the dynamic viscosity is obtained from:
        https://onlinelibrary.wiley.com/doi/pdf/10.1002/9780470516430.app2

        :return:
            The kinematic viscosity of air, measured in meters squared per second.

        """

        return self.dynamic_viscosity_of_air / self.density_of_air

    @property
    def sky_temperature(self) -> float:
        """
        Determines the radiative temperature of the sky.

        The "sky," as a black body, has a radiative temperature different to that of the
        surrounding air, or the ambient temperature. This function converts between them
        and outputs the sky's radiative temperature.

        :return:
            The radiative temperature of the "sky" in Kelvin.

        """

        return 0.0552 * (self.ambient_temperature ** 1.5)

    @property
    def thermal_conductivity_of_air(self) -> float:
        """
        The thermal conductivity of air varies as a function of temperature.

        The data for the thermal conductivity is obtained from:
        https://onlinelibrary.wiley.com/doi/pdf/10.1002/9780470516430.app2

        :return:
            The thermal conductivity of air, measured in Watts per meter Kelvin.

        """

        # This more accurate equation is not used by the paper.
        # return (0.02646 * self.ambient_temperature ** 1.5) / (
        #     self.ambient_temperature + 254.4 * (10 ** (-12 / self.ambient_temperature))
        # )

        # The reference suggests this equation is accurate to 1%.
        return 0.02646 * (self.ambient_temperature / 300) ** 0.8646

    @property
    def thermal_expansivity_of_air(self) -> float:
        """
        The thermal expansion coefficient of air varies as a function of temperature.

        The data for the thermal expansion coefficient is obtained from:
        https://onlinelibrary.wiley.com/doi/pdf/10.1002/9780470516430.app2

        :return:
            The thermal expansion coefficient of air, measured in Kelvin to the minus 1.

        """

        return 1 / self.ambient_temperature

    @property
    def wind_heat_transfer_coefficient(self) -> float:
        """
        Determines the convective heat transfer coefficient, either free, or forced.

        In the absence of any wind, the "free" wind_heat_transfer_coefficient is
        returned. If there is wind present, then this parameter is known as the "forced"
        wind_heat_transfer_coefficient.

        NOTE: The equation from Ilaria's paper is used in stead of Maria's.

        :return:
            The convective heat transfer coefficient due to the wind, measured in Watts
            per meter squared Kelvin.

        """

        return 3.8 + 2 * self.wind_speed
        # return 4.5 + 2.9 * self.wind_speed

    def __repr__(self) -> str:
        """
        Return a nice representation of the weather conditions.

        :return:
            A nicely-formatted string containing weather conditions data.

        """

        return (
            "WeatherConditions("
            f"ambient_temperature: {self.ambient_temperature:.3f}K, "
            f"azimuthal_angle: {self.azimuthal_angle}deg, "
            f"declination: {self.declination}deg, "
            f"density: {self.density_of_air:.3f}kg/m^3, "
            f"dynamic_viscosity: {self.dynamic_viscosity_of_air:.3f}kg/m*s, "
            f"heat_capacity: {self.heat_capacity_of_air}:.3fJ/kg*K, "
            f"irradiance: {self.irradiance:.3f}W/m^2, "
            f"kinematic_viscosity: {self.kinematic_viscosity_of_air:.3f}m^2/s, "
            f"sky_temperature: {self.sky_temperature:.3f}K, "
            f"thermal_conductivity: {self.thermal_conductivity_of_air:.3f}W/m*K, "
            f"thermal_expansion_coefficient: {self.thermal_expansivity_of_air:.3f}K^-1, "
            f"wind_heat_transfer_coefficient: {self.wind_heat_transfer_coefficient:2f}W/m*K, "
            f"wind_speed: {self.wind_speed:.3f}m/s, "
            ")"
        )


###################################
# PVT Layer Parameter Dataclasses #
###################################


@dataclass
class LayerParameters:
    """
    Contains parameters needed to instantiate a layer within the PV-T panel.

    .. attribute:: area
        The area of the layer, measured in meters squared.

    .. attribute:: conductivity
        The conductivity of the layer, measured in Watts per meter Kelvin.

    .. attribute:: density
        The density of the layer, measured in kilograms per meter cubed.

    .. attribute:: heat_capacity
        The heat capacity of the layer, measured in Joules per kilogram Kelvin.

    .. attribute:: mass
        The mass of the layer, measured in Kelvin.

    .. attribute:: thickness
        The thickness of the layer, measured in meters.

    """

    conductivity: float
    density: float
    heat_capacity: float
    thickness: float


@dataclass
class OpticalLayerParameters(LayerParameters):
    """
    Contains parameters needed to instantiate a layer with optical properties.

    .. attribute:: transmissivity
        The transmissivity of the layer: a dimensionless number between 0 (nothing is
        transmitted through the layer) and 1 (all light is transmitted).

    .. attribute:: absorptivity
        The absorptivity of the layer: a dimensionless number between 0 (nothing is
        absorbed by the layer) and 1 (all light is absorbed).

    .. attribute:: emissivity
        The emissivity of the layer; a dimensionless number between 0 (nothing is
        emitted by the layer) and 1 (the layer re-emits all incident light).

    """

    transmissivity: float
    absorptivity: float
    emissivity: float


@dataclass
class CollectorParameters(OpticalLayerParameters):
    """
    Contains parameters needed to instantiate a absorber layer within the PV-T panel.

    .. attribute:: htf_heat_capacity
        The heat capacity of the heat-transfer fluid through the absorber, measured in
        Joules per kilogram Kelvin.

    .. attribute:: inner_pipe_diameter
        The diameter of the inner wall of the pipes, in meters.

    .. attribute:: length
        The legnth of the absorber, measured in meters.

    .. attribute:: mass_flow_rate
        The mass flow rate of heat-transfer fluid through the absorber. Measured in
        litres per hour.

    .. attribute:: number_of_pipes
        The number of pipes attached to the back of the thermal absorber.
        NOTE: This parameter is very geography/design-specific, and will only be
        relevant/useful to the current design of absorber being modeled. Namely, when
        multiple pipes flow linearly down the length of the absorber, with the HTF
        taking a single pass through the absorber.

    .. attribute:: outer_pipe_diameter
        The diameter of the outer wall of the pipes, in meters.

    .. attribute:: pipe_density
        The density of the material making up the pipes, measured in kilograms per meter
        cubed.

    """

    htf_heat_capacity: float
    inner_pipe_diameter: float
    length: float
    mass_flow_rate: float
    number_of_pipes: int
    outer_pipe_diameter: float
    pipe_density: float


@dataclass
class PVParameters(OpticalLayerParameters):
    """
    Contains parameters needed to instantiate a PV layer within the PV-T panel.

    .. attribute:: reference_efficiency
        The efficiency of the PV layer at the reference temperature. Thie value varies
        between 1 (corresponding to 100% efficiency), and 0 (corresponding to 0%
        efficiency)

    .. attribute:: reference_temperature
        The referencee temperature, in Kelvin, at which the reference efficiency is
        defined.

    .. attribute:: thermal_coefficient
        The thermal coefficient for the efficiency of the panel.

    """

    reference_efficiency: float
    reference_temperature: float
    thermal_coefficient: float


####################
# Helper functions #
####################


def time_iterator(
    *,
    first_time: datetime.datetime,
    last_time: datetime.datetime,
    resolution: int,
    timezone: datetime.timezone,
) -> Generator[datetime.datetime, None, None]:
    """
    A generator function for looping through various times.

    :param first_time:
        The first time to be returned from the function.

    :param last_time:
        The last time, which, when reached, should cause the generator to stop.

    :param resolution:
        The time step, in seconds, for which the simulation should be run before saving.

    :param timezone:
        The timezone of the PV-T set-up.

    :return:
        A :class:`datetime.datetime` corresponding to the date and time at each point
        being itterated through.

    """

    current_time = first_time
    while current_time < last_time:
        yield current_time.replace(tzinfo=timezone)
        current_time += relativedelta(
            hours=resolution // 3600,
            minutes=(resolution // 60) % 60,
            seconds=resolution % 60,
        )



#!/usr/bin/python3.7
########################################################################################
# pvt_collector/pvt.py - Models a PVT panel and all contained components.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2020
########################################################################################

"""
The PV-T module for the PV-T model.

This module models the PV-T panel and its components, including the glass cover, PV
panel, and thermal absorber. The model works by incrimenting the parameters through at
increasing time steps, and the code here aims to emaulate this.

"""

import datetime
import math

from typing import Dict, Optional, Tuple

import numpy

from . import bond, absorber, glass, pv

from ...__utils__ import BColours, MissingParametersError, ProgrammerJudgementFault

from ..__utils__ import (
    CollectorParameters,
    OpticalLayerParameters,
    PVParameters,
    WeatherConditions,
)

from .__utils__ import MicroLayer
from .element import Element, ElementCoordinates

__all__ = ("PVT",)


# The minimum height in degrees that the sun must be above the horizon for light to
# strike the panel and be useful.
MINIMUM_SOLAR_DECLINATION = 5
# The maximum angle at which the panel can still be assumed to gain sunlight. This will
# limit both edge effects and prevent light reaching the panel from behind.
MAXIMUM_SOLAR_DIFF_ANGLE = 88


#####################
# PV-T Panel Layers #
#####################


class PVT:
    """
    Represents an entire PV-T absorber.

    .. attribute:: adhesive
        Represents the adhesive layer between the pv layer and the thermal absorber
        layer.

    .. attribute:: air_gap_thickness
        The thickness of the air gap between the glass and PV (or absorber) layers,
        measured in meters.

    .. attribute:: bond
        Represents the bond between the absorber (absorber) layer and the riser tube
        pipes.

    .. attribute:: absorber
        Represents the lower (thermal-absorber) layer of the panel.

    .. attribute:: eva
        Represents the EVA layer between the PV panel and the absorber layer.

    .. attribute:: glass
        Represents the upper (glass) layer of the panel. This is set to `None` if the
        panel is unglazed.

    .. attribute:: insulation
        Represents the thermal insulation on the back of the PV-T absorber.

    .. attribute:: latitude
        The latitude of the panel, measured in degrees.

    .. attribute:: longitude
        The longitude of the panel, measured in degrees.

    .. attribute:: portion_covered
        The portion of the PV-T panel which is covered with PV.

    .. attribute:: pv
         Represents the middle (pv) layer of the panel. Can be set to `None` if not
        present in the panel.

    .. attribute:: elements
        A mapping between element coordinates and the element.

    .. attribute:: tedlar
        Represents the tedlar back plate to the PV layer, situated above the PV layer.

    .. attribute:: timezone
        The timezone that the PVT system is based in.

    """

    # Private attributes:
    #
    # .. attribute:: _azimuthal_orientation
    #   The angle between the normal to the panel's surface and True North.
    #
    # .. attribute:: _horizontal_tracking
    #   A `bool` giving whether or not the panel tracks horizontally.
    #
    # .. attribute:: _tilt
    #   The angle between the normal to the panel's surface and the horizontal.
    #
    # .. attribute:: _vertical_tracking
    #   A `bool` giving whether or not the panel tracks verticallly.
    #

    def __init__(
        self,
        absorber_pipe_bond: bond.Bond,
        adhesive: MicroLayer,
        air_gap_thickness: Optional[float],
        area: float,
        absorber_parameters: CollectorParameters,
        eva: MicroLayer,
        glass_parameters: Optional[OpticalLayerParameters],
        insulation: MicroLayer,
        latitude: float,
        length: float,
        longitude: float,
        portion_covered: float,
        pv_parameters: PVParameters,
        elements: Dict[ElementCoordinates, Element],
        tedlar: MicroLayer,
        tilt: float,
        timezone: datetime.timezone,
        upper_glass_parameters: Optional[OpticalLayerParameters],
        width: float,
        *,
        azimuthal_orientation: Optional[float] = None,
        horizontal_tracking: bool = False,
        vertical_tracking: bool = False,
    ) -> None:
        """
        Instantiate an instance of the PV-T absorber class.

        :param adhesive:
            A :class:`MicroLayer` instance representing the adhesive layer.

        :param air_gap_thickness:
            The thickness, in meters, of the air gap between the PV and glass layers.
            `None` if there is no air gap present, e.g., if the glass layer is not
            present.

        :param area:
            The area of the panel, measured in meters squared.

        :param bond:
            A :class:`Bond` instance representing the bond layer.

        :param absorber_parameters:
            Parametsrs used to instantiate the absorber layer.

        :param eva:
            A :class:`MicroLayer` instance representing the eva layer.

        :param glass_parameters:
            Parameters used to instantiate the glass layer.

        :param insulation:
            A :class:`MicroLayer` instance representing the insulation on the back of
            the PV-T absorber.

        :param latitude:
            The latitude of the PV-T system, defined in degrees relative to the equator
            in the standard way.

        :param length:
            The length of the absorber in meters.

        :param longitude:
            The longitude of the PV-T system, defined in degrees relative to the
            Greenwich meridian.

        :param portion_covered:
            The portion of the PV-T panel which is covered with PV, ranging from 0
            (none) to 1 (all).

        :param pv_layer_included:
            Whether or not the system includes a photovoltaic layer.

        :param pv_parameters:
            Parameters used to instantiate the PV layer.

        :param elements:
            A mapping between element coordinate and element for all elements to be
            included in the layer.

        :param tedlar:
            A :class:`MicroLayer` instance representing the tedlar layer.

        :param timezone:
            The timezone in which the PV-T system is installed.

        :param upper_glass_parameters:
            Parameters used to instantiate the upper-glass layer.

        :param width:
            The width of the absorber in meters.

        :param azimuthal_orientation:
            The orientation of the surface of the panel, defined relative to True North,
            measured in degrees.

        :param horizontal_tracking:
            Whether or not the PV-T system tracks the sun horizontally.

        :param tilt:
            The tilt of the panel above the horizontal, measured in degrees.

        :param vertical_tracking:
            Whether or not the PV-T system tracks the sun vertically.

        :raises: MissingParametesrError
            Raised if not all the parameters needed are supplied.

        """

        self._azimuthal_orientation = azimuthal_orientation
        self._horizontal_tracking = horizontal_tracking
        self._tilt = tilt
        self._vertical_tracking = vertical_tracking
        self.air_gap_thickness = air_gap_thickness
        self.area = area
        self.latitude = latitude
        self.length = length
        self.longitude = longitude
        self.portion_covered = portion_covered
        self.elements = elements
        self.timezone = timezone
        self.width = width

        # Set up empty placeholders to be computed later.
        self._absorber_transmissivity_absorptivity_product: Optional[float] = None
        self._glass_transmissivity_absorptivity_product: Optional[float] = None
        self._pv_transmissivity_absorptivity_product: Optional[float] = None
        self._upper_glass_transmissivity_absorptivity_product: Optional[float] = None

        # If the orientation parameters have not been specified correctly, then raise an
        # error.
        if not vertical_tracking and tilt is None:
            raise MissingParametersError(
                "PVT",
                "If the panel does not track vertically, then the tilt must be given.",
            )
        if not horizontal_tracking and azimuthal_orientation is None:
            raise MissingParametersError(
                "PVT",
                "If the panel does not track horizontally, then the azimuthal "
                "orientation must be given.",
            )

        # Instantiate the layers.
        self.adhesive = adhesive
        self.bond = absorber_pipe_bond
        self.absorber = absorber.Collector(absorber_parameters)
        self.eva = eva
        if glass_parameters is None:
            self.glass: Optional[glass.Glass] = None
        else:
            self.glass = glass.Glass(glass_parameters)
        self.insulation = insulation
        self.pv: pv.PV = pv.PV(pv_parameters)
        self.tedlar = tedlar
        if upper_glass_parameters is None:
            self.upper_glass: Optional[glass.Glass] = None
        else:
            self.upper_glass = glass.Glass(upper_glass_parameters)

        # * Instantiate and store the elements on the class.

    def __repr__(self) -> str:
        """
        A nice-looking representation of the PV-T panel.

        :return:
            A nice-looking representation of the PV-T panel.

        """

        return (
            "PVT(\n"
            f"  upper_glass: {self.upper_glass},\n"
            f"  glass: {self.glass},\n"
            f"  pv: {self.pv},\n"
            f"  eva: {self.eva},\n"
            f"  adhesive: {self.adhesive}\n"
            f"  tedlar: {self.tedlar}\n"
            f"  absorber: {self.absorber},\n"
            f"  bond: {self.bond},\n"
            f"  outer_pipe_diameter: {self.absorber.outer_pipe_diameter}m\n"
            f"  (ta)_ug = {self.upper_glass_transmissivity_absorptivity_product:.3f}\n"
            f"  (ta)_g = {self.glass_transmissivity_absorptivity_product:.3f}\n"
            f"  (ta)_pv = {self.pv_transmissivity_absorptivity_product:.3f}\n"
            f"  (ta)_ab = {self.absorber_transmissivity_absorptivity_product:.3f}\n"
            "  additional_params: "
            f"azimuthal_orientation: {self._azimuthal_orientation}, "
            f"coordinates: {self.latitude}N {self.longitude}E, "
            f"tilt: {self._tilt}deg"
            "\n)"
        )

    @property
    def _double_glass_reflectance(self) -> float:
        """
        Computes the combined reflectance of two glass layers.

        :return:
            The dimensionless reflectance of the combined upper-glass and glass layers.

        """

        if self.upper_glass is None or self.glass is None:
            raise ProgrammerJudgementFault(
                "{}Cannot compute combined reflectances without both ".format(
                    BColours.FAIL
                )
                + "upper-glass and glass layers present.{}".format(BColours.ENDC)
            )

        return self.upper_glass.reflectance + (
            self.glass.reflectance * self.upper_glass.transmittance ** 2
        ) / (1 - self.upper_glass.reflectance * self.glass.reflectance)

    @property
    def _double_glass_transmittance(self) -> float:
        """
        Computes the combined transmittance of two glass layers.

        :return:
            The dimensionless transmittance of the combined upper-glass and glass layers.

        """

        if self.upper_glass is None or self.glass is None:
            raise ProgrammerJudgementFault(
                "{}Cannot compute combined transmittances without both ".format(
                    BColours.FAIL
                )
                + "upper-glass and glass layers present.{}".format(BColours.ENDC)
            )

        return (
            self.upper_glass.transmittance
            * self.glass.transmittance
            / (1 - self.upper_glass.reflectance * self.glass.reflectance)
        )

    def _get_solar_orientation_diff(
        self, azimuthal_angle: float, declination: float
    ) -> float:
        """
        Determine the between the panel's normal and the sun.

        :param azimuthal_angle:
            The position of the sun, defined as degrees clockwise from True North.

        :param declination:
            The declination of the sun: the angle in degrees of the sun above the
            horizon.

        :return:
            The angle in degrees between the solar irradiance and the normal to the
            panel.

        """

        # Determine the angle between the sun and panel's normal, both horizontally and
        # vertically. If tracking is enabled, then this angle should be zero along each
        # of those axes. If tracking is disabled, then this angle is just the difference
        # between the panel's orientation and the sun's.

        if self._horizontal_tracking or self._azimuthal_orientation is None:
            horizontal_diff: float = 0
        else:
            horizontal_diff = abs(self._azimuthal_orientation - azimuthal_angle)

        if self._vertical_tracking or self._tilt is None:
            vertical_diff: float = 0
        else:
            vertical_diff = abs(self._tilt - declination)

        # Combine these to generate the angle between the two directions.
        return math.degrees(
            math.acos(
                math.cos(math.radians(horizontal_diff))
                * math.cos(math.radians(vertical_diff))
            )
        )

    @property
    def absorber_transmissivity_absorptivity_product(self) -> float:
        """
        Returns the transmissivity-absorptivity product of the Absorber layer.

        Due to internal reflections etc., an estimate is needed for the overall fraction
        of incident solar light that is absorbed by the Absorber layer. NOTE: This only
        applies when there is no PV layer present, and light can pass straight through
        the glass layer onto the solar-thermal absorber layer.

        :return:
            The TA product of the Absorber layer.

        """

        if self._absorber_transmissivity_absorptivity_product is not None:
            return self._absorber_transmissivity_absorptivity_product

        if self.upper_glass is not None:
            self._absorber_transmissivity_absorptivity_product = (
                (1 - self.absorber.reflectance - self.absorber.transmittance)
                * self._double_glass_transmittance
                / (1 - self._double_glass_reflectance * self.absorber.reflectance)
            )
        elif self.glass is not None:
            self._absorber_transmissivity_absorptivity_product = (
                (1 - self.absorber.reflectance - self.absorber.transmittance)
                * self.glass.transmittance
            ) / (1 - self.absorber.reflectance * self.glass.reflectance)
        else:
            self._absorber_transmissivity_absorptivity_product = (
                self.absorber.absorptivity
            )

        return self._absorber_transmissivity_absorptivity_product

    @property
    def coordinates(self) -> Tuple[float, float]:
        """
        Returns a the coordinates of the panel.

        :return:
            A `tuple` containing the latitude and longitude of the panel.

        """

        return (self.latitude, self.longitude)

    def electrical_output(
        self, pv_temperature: float, weather_conditions: WeatherConditions
    ) -> float:
        """
        Returns the electrical output of the PV-T panel in Watts.

        NOTE: We here need to include the portion of the panel that is covered s.t. the
        correct electricitiy-generating area is accounted for, rather than accidentailly
        inculding areas which do not generated electricity.

        :param pv_temperature:
            The temperature of the PV layer, measured in Kelvin.

        :param weather_conditions:
            The current weather conditions at the time step being incremented to.

        :return:
            The electrical output, in Watts, of the PV-T panel.

        """

        if self.pv is not None:
            electrical_output: float = (
                self.pv.electrical_efficiency(pv_temperature)
                * self.get_solar_irradiance(weather_conditions)
                * self.area
                * self.portion_covered
            )
        else:
            electrical_output = 0

        return electrical_output

    def get_solar_irradiance(  # pylint: disable=no-self-use
        self, weather_conditions: WeatherConditions
    ) -> float:
        """
        Returns the irradiance striking the panel normally in Watts per meter squared.

        :param weather_conditions:
            The current weather conditions.

        :return:
            The solar irradiance striking the panel, adjusted for the sun's declination
            and azimuthal angle, as well as the panel's orientation, dependant on the
            panel's latitude and longitude.

        """

        # >>> Currently, Maria's profiles, rather than my Maths, are used.
        # # Compute the angle of solar irradiance wrt the panel.
        # solar_diff_angle = self._get_solar_orientation_diff(
        #     weather_conditions.azimuthal_angle,
        #     weather_conditions.declination,
        # )

        # return (
        #     weather_conditions.irradiance  # [W/m^2]
        #     * math.cos(math.radians(solar_diff_angle))
        #     if weather_conditions.declination >= MINIMUM_SOLAR_DECLINATION
        #     and 0 <= solar_diff_angle <= MAXIMUM_SOLAR_DIFF_ANGLE
        #     else 0
        # )
        # <<< End of my Maths.
        # >>> Beginning of Maria's profile fetching.
        return weather_conditions.irradiance
        # <<< End of Maria's profile fetching.

    @property
    def glass_transmissivity_absorptivity_product(self) -> float:
        """
        Returns the transmissivity-absorptivity product of the glass layer.

        Due to internal reflections etc., an estimate is needed for the overall fraction
        of incident solar light that is absorbed by the glass layer.

        :return:
            The TA product of the glass layer.

        """

        if self._glass_transmissivity_absorptivity_product is not None:
            return self._glass_transmissivity_absorptivity_product

        if self.glass is None:
            return 0

        if self.pv is not None:
            lower_reflectance = self.pv.reflectance
        else:
            lower_reflectance = self.absorber.reflectance

        if self.upper_glass is None:
            self._glass_transmissivity_absorptivity_product = (
                1
                - self.glass.reflectance
                - self.glass.transmittance
                * (1 - lower_reflectance + lower_reflectance * self.glass.transmittance)
                / (1 - lower_reflectance * self.glass.reflectance)
            )
        else:
            self._glass_transmissivity_absorptivity_product = (
                1
                - self._double_glass_reflectance
                - (
                    1
                    - lower_reflectance
                    + lower_reflectance * self._double_glass_transmittance
                )
                * self._double_glass_transmittance
                / (1 - self._double_glass_reflectance * lower_reflectance)
            )

        return self._glass_transmissivity_absorptivity_product

    @property
    def pv_to_absorber_thermal_resistance(self) -> float:
        """
        Returns the thermal resistance between the PV and absorber layers.

        :return:
            The thermal resistance, in meters squared Kelvin per Watt, between the PV
            and absorber layers.

        """

        return (
            self.eva.thickness / self.eva.conductivity
            + self.tedlar.thickness / self.tedlar.conductivity
            + self.adhesive.thickness / self.adhesive.conductivity
        )

    @property
    def pv_transmissivity_absorptivity_product(self) -> float:
        """
        Returns the transmissivity-absorptivity product of the PV layer.

        Due to internal reflections etc., an estimate is needed for the overall fraction
        of incident solar light that is absorbed by the PV layer.

        :return:
            The TA product of the PV layer.

        """

        if self._pv_transmissivity_absorptivity_product is not None:
            return self._pv_transmissivity_absorptivity_product

        if self.upper_glass is not None:
            self._pv_transmissivity_absorptivity_product = (
                (1 - self.pv.reflectance - self.pv.transmittance)
                * self._double_glass_transmittance
                / (1 - self._double_glass_reflectance * self.pv.reflectance)
            )
        elif self.glass is not None:
            # if self.glass is not None:
            self._pv_transmissivity_absorptivity_product = (
                (1 - self.pv.reflectance - self.pv.transmittance)
                * self.glass.transmittance
            ) / (1 - self.pv.reflectance * self.glass.reflectance)
        else:
            self._pv_transmissivity_absorptivity_product = self.pv.absorptivity

        return self._pv_transmissivity_absorptivity_product

    @property
    def tilt_in_radians(self) -> float:
        """
        Returns the tilt of the panel from the horizontal, measured in radians.

        :return:
            The angle between the plane of the panel and the horizontal, measured in
            radians.

        """

        return numpy.pi * self._tilt / 180

    @property
    def upper_glass_transmissivity_absorptivity_product(self) -> float:
        """
        Returns the transmissivity-absorptivity product of the upper-glass layer.

        Due to internal reflections etc., an estimate is needed for the overall fraction
        of incident solar light that is absorbed by the upper-glass layer.

        :return:
            The TA product of the upper-glass layer.

        """

        if self._upper_glass_transmissivity_absorptivity_product is not None:
            return self._upper_glass_transmissivity_absorptivity_product

        if self.upper_glass is None:
            return 0

        if self.glass is None:
            raise ProgrammerJudgementFault(
                "{}Attempted to fetch ta product of a double glazing without a ".format(
                    BColours.FAIL
                )
                + "lower glass layer.{}".format(BColours.ENDC)
            )

        lower_reflectance = self.glass.reflectance

        self._upper_glass_transmissivity_absorptivity_product = (
            1
            - self.upper_glass.reflectance
            - (
                1
                - lower_reflectance
                + lower_reflectance * self.upper_glass.transmittance
            )
            * self.upper_glass.transmittance
            / (1 - self.upper_glass.reflectance * lower_reflectance)
        )

        return self._upper_glass_transmissivity_absorptivity_product



#!/usr/bin/python3.7
# type: ignore
########################################################################################
# thesis_analysis.py - Carries out analysis on data available from Ilaria in her thesis.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################
"""
Carries out analysis to compare results produced by my model to Ilaria's thesis.

The glazing validation, analysed in `glazing_analysis.py`, has produced results that
differ from those obtained by Ilaria with her model. However, in her thesis, she has
experimental results that look more similar to those obtained by my model.

This module hence carries out analysis to plot my results against those obtained
experimentally by Ilaria for comparison and validation purposes.

"""

import argparse
import collections
import os
import sys

from logging import Logger
from typing import Any, Dict, Optional, Union

import re
import numpy
import yaml

from matplotlib import pyplot as plt

try:
    from ..__utils__ import get_logger
    from .__utils__ import (
        GraphDetail,
        load_model_data,
        plot_figure,
    )
except ModuleNotFoundError:
    import logging

    logging.error(
        "Incorrect module import. Try running with `python3.7 -m pvt_model.analysis`"
    )
    raise

__all__ = ("analyse",)

# Used to distinguish copled data sets.
COUPLED_DATA_TYPE = "coupled"
# Used to distinguished decoupled data sets.
DECOUPLED_DATA_TYPE = "decoupled"
# Used to identify the "time" data set.
TIME_KEY = "time"
# Used to distinguish steady-state data sets.
STEADY_STATE_DATA_TYPE = "steady_state"
# Name of the steady-state data file.
STEADY_STATE_DATA_FILE_NAME = "ilaria_glazing_validation_runs.yaml"
# Height of the y error bars.
Y_ERROR_BAR_HEIGHT = 0.1


def _parse_args(args) -> argparse.Namespace:
    """
    Parse the CLI args.

    """

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--data-file-directory",
        "-dfdir",
        help="Path to the directory containing data files to parse.",
    )
    parser.add_argument(
        "--show-output",
        "-so",
        action="store_true",
        default=False,
        help="Show the output figures generated.",
    )
    parser.add_argument(
        "--skip-2d-plots",
        "-skip",
        action="store_true",
        default=False,
        help="Skip plotting of 2D figures.",
    )

    return parser.parse_args(args)


def _resolution_from_graph_detail(
    graph_detail: GraphDetail, num_data_points: int
) -> int:
    """
    Determine the x-step resolution of the plot.

    I realise here that the word "resolution" is over-used. Here, the number of data
    points (in the system data) that need to be absorbed into one graph point is
    calculated and returned.

    :param graph_detail:
        The detail level needed on the graph.

    :param num_data_points:
        The number of data points recorded in the plot.

    :return:
        The number of data points that need to be absorbed per graph point.

    """

    # * For "lowest", include one point every half-hour.
    if graph_detail == GraphDetail.highest:
        return 1

    return int(num_data_points / graph_detail.value)


def _reduce_data(  # pylint: disable=too-many-branches
    data_to_reduce: Dict[str, Dict[Any, Any]],
    graph_detail: GraphDetail,
    logger: Logger,
) -> Dict[Union[int, str], Dict[Any, Any]]:
    """
    This processes the data, using sums to reduce the resolution so it can be plotted.

    :param data_to_reduce:
        The raw, JSON data, contained within a `dict`.

    :param graph_detail:
        The level of detail required in the graph.

    :param logger:
        The logger being used in the run.

    :return:
        The cropped/summed up data, returned at a lower resolution as specified by the
        graph detail.

    """

    logger.info("Beginning data reduction.")

    # Determine the number of data points to be amalgamated per graph point.
    data_points_per_graph_point: int = _resolution_from_graph_detail(
        graph_detail, len(data_to_reduce)
    )

    logger.info(
        "%s data points will be condensed to each graph point.",
        data_points_per_graph_point,
    )

    if data_points_per_graph_point <= 1:
        return data_to_reduce

    # Construct a dictionary to contain this reduced data.
    reduced_data: Dict[Union[int, str], Dict[Any, Any]] = {
        index: dict()
        for index in range(
            int(
                len([key for key in data_to_reduce if key.isdigit()])
                / data_points_per_graph_point
            )
        )
    }

    # Depending on the type of data entry, i.e., whether it is a temperature, load,
    # demand covered, or irradiance (or other), the way that it is processed will vary.
    for data_entry_name in data_to_reduce[  # pylint: disable=too-many-nested-blocks
        "0"
    ].keys():
        # pdb.set_trace(header="Beginning of reduction loop.")
        # * If the entry is a date or time, just take the value
        if data_entry_name in ["date", TIME_KEY]:
            for index, _ in enumerate(reduced_data):
                reduced_data[index][data_entry_name] = data_to_reduce[
                    str(index * data_points_per_graph_point)
                ][data_entry_name]
            continue

        # * If the data entry is a temperature, or a power output, then take a rolling
        # average
        if any(
            (
                key in data_entry_name
                for key in [
                    "temperature",
                    "irradiance",
                    "efficiency",
                    "hot_water_load",
                    "electrical_load",
                ]
            )
        ):
            for outer_index, _ in enumerate(reduced_data):
                # Attempt to process as a dict or float first.
                if isinstance(data_to_reduce["0"][data_entry_name], float):
                    reduced_data[outer_index][data_entry_name] = sum(
                        [
                            float(data_to_reduce[str(inner_index)][data_entry_name])
                            / data_points_per_graph_point
                            for inner_index in range(
                                int(data_points_per_graph_point * outer_index),
                                int(data_points_per_graph_point * (outer_index + 1)),
                            )
                        ]
                    )
                elif isinstance(data_to_reduce["0"][data_entry_name], dict):
                    # Loop through the various coordinate pairs.
                    for sub_dict_key in data_to_reduce[str(outer_index)][
                        data_entry_name
                    ]:
                        if data_entry_name not in reduced_data[outer_index]:
                            reduced_data[outer_index][data_entry_name]: Dict[
                                str, float
                            ] = dict()
                        reduced_data[outer_index][data_entry_name][sub_dict_key] = sum(
                            [
                                float(
                                    data_to_reduce[str(inner_index)][data_entry_name][
                                        sub_dict_key
                                    ]
                                )
                                / data_points_per_graph_point
                                for inner_index in range(
                                    int(data_points_per_graph_point * outer_index),
                                    int(
                                        data_points_per_graph_point * (outer_index + 1)
                                    ),
                                )
                            ]
                        )
                else:
                    logger.debug(
                        "A value was an unsuported type. Setting to 'None': %s",
                        data_entry_name,
                    )
                    for index, _ in enumerate(reduced_data):
                        reduced_data[index][data_entry_name] = None
                    continue

        # * If the data entry is a load, then take a sum
        elif any((key in data_entry_name for key in {"load", "output"})):
            # @@@
            # * Here, the data is divided by 3600 to convert from Joules to Watt Hours.
            # * This only works provided that we are dealing with values in Joules...
            for outer_index, _ in enumerate(reduced_data):
                reduced_data[outer_index][data_entry_name] = (
                    sum(
                        [
                            float(data_to_reduce[str(inner_index)][data_entry_name])
                            for inner_index in range(
                                int(data_points_per_graph_point * outer_index),
                                int(data_points_per_graph_point * (outer_index + 1)),
                            )
                        ]
                    )
                    / 3600
                )
            continue

        # * Otherwise, we just use the data point.
        for index, _ in enumerate(reduced_data):
            reduced_data[index][data_entry_name] = data_to_reduce[
                str(index * data_points_per_graph_point)
            ][data_entry_name]
        continue

    return reduced_data


def _post_process_data(
    data_to_post_process: Dict[str, Dict[Any, Any]]
) -> Dict[str, Dict[Any, Any]]:
    """
    Carries out post-processing on data where necessary.

    Things to be computed:
        - Bulk Water Temperautre
        - Litres consumed

    :param data_to_post_process:
        The data to be post processed.

    :return:
        The post-processed data.

    """

    # * Cycle through all the data points and compute the new values as needed.
    for data_entry in data_to_post_process.values():
        data_entry["absorber_temperature_gain"] = (
            data_entry["collector_output_temperature"]
            - data_entry["collector_input_temperature"]
        )
    #     # Conversion needed from Wh to Joules.
    #     data_entry["litres_per_hour"] = (
    #         data_entry["thermal_load"] / (HEAT_CAPACITY_OF_WATER * 50) * 60
    #     )
    return data_to_post_process


def _plot_experimental_data() -> plt.Axes:
    """
    Plot the experimental data for the runs, both unglazed and single glazed.

    :return:
        The axes used for the plotting.

    """

    # Parse the thermal-efficiency data.
    with open(
        os.path.join("validation_data", STEADY_STATE_DATA_FILE_NAME),
        "r",
    ) as f:  #
        experimental_steady_state_data = yaml.safe_load(f)

    # Plot the experimental data.
    _, ax1 = plt.subplots()
    for glazing, colour in zip(["unglazed", "single_glazed"], ["blue", "orange"]):
        x_data = [
            entry["reduced_temperature"]
            for entry in experimental_steady_state_data[glazing]
        ]
        y_data = [
            entry["thermal_efficiency"]
            for entry in experimental_steady_state_data[glazing]
        ]
        error_bar_data = [
            entry["error_bar"] for entry in experimental_steady_state_data[glazing]
        ]
        ax1.scatter(
            x_data,
            y_data,
            # yerr=Y_ERROR_BAR_HEIGHT,
            marker="s",
            # ls="none",
        )
        plt.errorbar(
            x_data,
            y_data,
            yerr=error_bar_data,
            ls="none",
            color=colour,
        )
        fit = numpy.polyfit(x_data, y_data, 2)
        x_series = numpy.linspace(min(x_data), max(x_data), 100)
        y_series = [fit[0] * entry ** 2 + fit[1] * entry + fit[2] for entry in x_series]
        plt.plot(x_series, y_series)

    return ax1


def analyse_decoupled_steady_state_data(
    data: Dict[Any, Any], logger: Logger, skip_2d_plots: bool
) -> None:
    """
    Carry out analysis on a set of steady-state data.

    :param data:
        The data to analyse.

    :param logger:
        The logger to use for the analysis run.

    :param skip_2d_plots:
        Whether to skip the 2D plots (True) or plot them (False).

    """

    logger.info("Beginning steady-state analysis.")

    if skip_2d_plots:
        print(f"{int(len(data.keys()) * 2 + 2)} figures will be plotted.")
        logger.info("%s figures will be plotted.", int(len(data.keys()) * 2 + 2))
    else:
        print(f"{int(len(data.keys()) * 5 + 2)} figures will be plotted.")
        logger.info("%s figures will be plotted.", int(len(data.keys()) * 5 + 2))

    # Post-process the data to a plottable format.
    single_glazed_data = collections.defaultdict(dict)
    unglazed_data = collections.defaultdict(dict)
    single_glazed_thermal_efficiency_labels = set()
    unglazed_thermal_efficiency_labels = set()
    ilaria_glazing_regex = re.compile(r"ilaria_(?P<glazing>.*)_steady_state_runs.*")
    for key, sub_dict in data.items():
        ilaria_glazing_match = re.match(ilaria_glazing_regex, key)
        if ilaria_glazing_match is None:
            continue
        glazing = ilaria_glazing_match.group("glazing")
        if glazing == "single_glazed":
            for sub_key, value in sub_dict.items():
                single_glazed_data[sub_key][f"{glazing} thermal efficiency"] = value[
                    "thermal_efficiency"
                ]
                single_glazed_thermal_efficiency_labels.add(
                    f"{glazing} thermal efficiency"
                )
                # Store a copy of the reduced temperature
                single_glazed_data[sub_key]["reduced_collector_temperature"] = value[
                    "reduced_collector_temperature"
                ]
        else:
            for sub_key, value in sub_dict.items():
                unglazed_data[sub_key][f"{glazing} thermal efficiency"] = value[
                    "thermal_efficiency"
                ]
                unglazed_thermal_efficiency_labels.add(f"{glazing} thermal efficiency")
                # Store a copy of the reduced temperature
                unglazed_data[sub_key]["reduced_collector_temperature"] = value[
                    "reduced_collector_temperature"
                ]

    # Thermal efficiency plot.
    logger.info("Plotting thermal efficiency against the reduced temperature.")

    # Plot the experimental data.
    ax1 = _plot_experimental_data()

    # Add the model data.
    plot_figure(
        "thesis_glazing_analysis_thermal_efficiency_against_reduced_temperature",
        single_glazed_data,
        first_axis_things_to_plot=single_glazed_thermal_efficiency_labels,
        first_axis_label="Thermal efficiency",
        x_axis_label="Reduced temperature / K m^2 / W",
        x_axis_thing_to_plot="reduced_collector_temperature",
        plot_title="Thermal efficiency against reduced temperature",
        disable_lines=True,
        plot_trendline=True,
        override_axis=ax1,
        first_axis_y_limits=[-0.1, 0.6],
        transparent=True,
    )
    plot_figure(
        "thesis_glazing_analysis_thermal_efficiency_against_reduced_temperature",
        unglazed_data,
        first_axis_things_to_plot=unglazed_thermal_efficiency_labels,
        first_axis_label="Thermal efficiency",
        x_axis_label="Reduced temperature / K m^2 / W",
        x_axis_thing_to_plot="reduced_collector_temperature",
        plot_title="Thermal efficiency against reduced temperature",
        disable_lines=True,
        plot_trendline=True,
        override_axis=ax1,
        first_axis_y_limits=[-0.1, 0.6],
        transparent=True,
    )


def analyse(
    data_file_directory: str,
    show_output: Optional[bool] = False,
    skip_2d_plots: Optional[bool] = False,
) -> None:
    """
    The main method for the analysis module.

    :param data_file_directory:
        The path to the directory containing data files to analyse.

    :param show_output:
        Whether to show the output files generated.

    :param skip_2d_plots:
        Whether to skip the 2D plots (True) of include them (False).

    """

    # * Set up the logger
    logger = get_logger(True, "pvt_thesis_analysis", True)

    # * Extract the data.
    data: Dict[str, Dict[str, Any]] = dict()
    for data_file_name in os.listdir(data_file_directory):
        data[data_file_name] = load_model_data(
            os.path.join(data_file_directory, data_file_name)
        )

        # If the data type is not decoupled, steady-state data, then exit.
        data_type = data[data_file_name].pop("data_type")
        if data_type != f"{DECOUPLED_DATA_TYPE}_{STEADY_STATE_DATA_TYPE}":
            logger.info("Data type was not 'steady_state'. Omitting...")
            data.pop(data_file_name)
            continue
        if "no_pv" in data_file_name:
            logger.info("Invalid data was removed.")
            data.pop(data_file_name)
            continue

    analyse_decoupled_steady_state_data(data, logger, skip_2d_plots)

    logger.info("Analysis complete - all figures saved successfully.")

    if show_output:
        plt.show()


if __name__ == "__main__":
    parsed_args = _parse_args(sys.argv[1:])
    analyse(
        parsed_args.data_file_directory,
        parsed_args.show_output,
        parsed_args.skip_2d_plots,
    )



#!/usr/bin/python3.7
########################################################################################
# __main__.py - The main module for the higher-level PV-T model.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The main module for the PV-T model.

The model is run from here for several runs as determined by command-line arguments.

"""

import os
import sys

from argparse import Namespace
from functools import partial
from logging import Logger
from multiprocessing import Pool
from statistics import mean
from typing import Any, Dict, List, Optional, Set, Tuple

from tqdm import tqdm

from . import argparser

from .__utils__ import (
    BColours,
    COARSE_RUN_RESOLUTION,
    FileType,
    SteadyStateRun,
    fourier_number,
    get_logger,
    INITIAL_CONDITION_PRECISION,
    LOGGER_NAME,
    OperatingMode,
    ProgrammerJudgementFault,
    read_yaml,
    save_data,
    SystemData,
    TemperatureName,
)

from .analysis import analysis

from .pvt_system import index_handler, tank
from .pvt_system.pvt_collector import pvt
from .pvt_system.__main__ import main as pvt_system_main
from .pvt_system.__utils__ import DivergentSolutionError

from .pvt_system.constants import (
    DEFAULT_SYSTEM_TEMPERATURE,
    DENSITY_OF_WATER,
    HEAT_CAPACITY_OF_WATER,
    THERMAL_CONDUCTIVITY_OF_WATER,
    ZERO_CELCIUS_OFFSET,
)
from .pvt_system.process_pvt_system_data import (
    hot_water_tank_from_path,
    pvt_collector_from_path,
)


# Done message:
#   The message to display when a task was successful.
DONE: str = "[   DONE   ]"

# Failed message:
#   The message to display when a task has failed.
FAILED: str = "[  FAILED  ]"

# PVT header string:
#   Text to display when instantiating the PV-T model.
PVT_HEADER_STRING = """

                                       ,
                                       ,
                          ,           ,,,           ,
                           ,,         ,,,         ,,
                            ,,,.     .,,,,      ,,,
                             ,,,,.            ,,,,.
                 ,,.              ,,,,,,,,,,,    ,         .,,.
                    ,,,,,,.   ,,,,,,,,,,,,,,,,,,,   .,,,,,,.
                       ,,,  ,,,,,,,,,,,,,,,,,,,,,,,  ,,,.
                           ,,,,,,,,,,,,,,,,,,,,,,,,,
                          ,,,,,,,,,,,,,,,,,,,,,,,,,,,
           .,,,,,,,,,,,,  ,,,,,,,,,.,.,,,,,,,,,,,,,,,  ,,,,,,,,,,,,,

               _    _ ______       _______ _____                 _
              | |  | |  ____|   /\\|__   __|  __ \\               | |
              | |__| | |__     /  \\  | |  | |__) |_ _ _ __   ___| |
              |  __  |  __|   / /\\ \\ | |  |  ___/ _` | '_ \\ / _ \\ |
              | |  | | |____ / ____ \\| |  | |  | (_| | | | |  __/ |
              |_|  |_|______/_/    \\_\\_|  |_|   \\__,_|_| |_|\\___|_|


                    Hybrid Electric And Thermal Panel Model
                      Copyright Benedict Winchester, 2021

              For more information, contact Benedict Winchester at
                         benedict.winchester@gmail.com

"""


def _get_system_fourier_numbers(
    hot_water_tank: Optional[tank.Tank], pvt_collector: pvt.PVT, resolution: float
) -> Dict[TemperatureName, float]:
    """
    Determine the Fourier numbers of the various system components.

    :param hot_water_tank:
        A :class:`tank.Tank` instance representing the hot-water tank in the system.

    :param pvt_collector:
        A :class:`pvt.PVT` instance representing the PVT panel being modelled.

    :param resolution:
        The resolution being used for the model.

    :return:
        The Fourier coefficients for the various components being modelled.

    """

    # Determine the Fourier coefficients of the panel's layers.
    fourier_number_map: Dict[TemperatureName, float] = dict()
    if pvt_collector.glass is not None:
        fourier_number_map[TemperatureName.glass] = round(
            fourier_number(
                pvt_collector.glass.thickness,
                pvt_collector.glass.conductivity,
                pvt_collector.glass.density,
                pvt_collector.glass.heat_capacity,
                resolution,
            ),
            2,
        )
    if pvt_collector.pv is not None:
        fourier_number_map[TemperatureName.pv] = round(
            fourier_number(
                pvt_collector.pv.thickness,
                pvt_collector.pv.conductivity,
                pvt_collector.pv.density,
                pvt_collector.pv.heat_capacity,
                resolution,
            ),
            2,
        )
    fourier_number_map[TemperatureName.absorber] = round(
        fourier_number(
            pvt_collector.absorber.thickness,
            pvt_collector.absorber.conductivity,
            pvt_collector.absorber.density,
            pvt_collector.absorber.heat_capacity,
            resolution,
        ),
        2,
    )
    fourier_number_map[TemperatureName.htf] = round(
        fourier_number(
            pvt_collector.absorber.inner_pipe_diameter,
            THERMAL_CONDUCTIVITY_OF_WATER,
            DENSITY_OF_WATER,
            pvt_collector.absorber.htf_heat_capacity,
            resolution,
        ),
        2,
    )
    if hot_water_tank is not None:
        fourier_number_map[TemperatureName.tank] = round(
            fourier_number(
                hot_water_tank.diameter,
                THERMAL_CONDUCTIVITY_OF_WATER,
                DENSITY_OF_WATER,
                HEAT_CAPACITY_OF_WATER,
                resolution,
            ),
            5,
        )

    return fourier_number_map


def _calculate_and_print_fourier_numbers(
    hot_water_tank: Optional[tank.Tank],
    logger: Logger,
    parsed_args: Namespace,
    pvt_collector: pvt.PVT,
) -> None:
    """
    Determines, prints, and logs the various Fourier numbers.

    :param hot_water_tank:
        A :class:`tank.Tank` instance representing the hot-water tank being modelled.

    :param logger:
        The logger being used for the run.

    :param parsed_args:
        The parsed commnand-line arguments.

    :param pvt_collector:
        The pvt panel, represented as a :class:`pvt_collector.pvt.PVT` instance.

    """

    # Determine the Fourier numbers.
    fourier_number_map = _get_system_fourier_numbers(
        hot_water_tank, pvt_collector, parsed_args.resolution
    )

    logger.info(
        "Fourier numbers determined:\n%s\n%s",
        "|".join(
            [
                " {}{}".format(
                    key.name,
                    " "
                    * (
                        max(
                            [len(key.name) for key in fourier_number_map]
                            + [len(str(value)) for value in fourier_number_map.values()]
                        )
                        - len(key.name)
                        + 1
                    ),
                )
                for key in fourier_number_map
            ]
        ),
        "|".join(
            [
                " {}{}".format(
                    value,
                    " "
                    * (
                        max(
                            [len(key.name) for key in fourier_number_map]
                            + [len(str(value)) for value in fourier_number_map.values()]
                        )
                        - len(str(value))
                        + 1
                    ),
                )
                for value in fourier_number_map.values()
            ]
        ),
    )
    print(
        "Fourier numbers determined:\n{}\n{}".format(
            "|".join(
                [
                    " {}{}".format(
                        key.name,
                        " "
                        * (
                            max(
                                [len(key.name) for key in fourier_number_map]
                                + [
                                    len(str(value))
                                    for value in fourier_number_map.values()
                                ]
                            )
                            - len(key.name)
                            + 1
                        ),
                    )
                    for key in fourier_number_map
                ]
            ),
            "|".join(
                [
                    " {}{}".format(
                        value,
                        " "
                        * (
                            max(
                                [len(key.name) for key in fourier_number_map]
                                + [
                                    len(str(value))
                                    for value in fourier_number_map.values()
                                ]
                            )
                            - len(str(value))
                            + 1
                        ),
                    )
                    for value in fourier_number_map.values()
                ]
            ),
        )
    )


def _determine_consistent_conditions(
    number_of_pipes: int,
    layers: Set[TemperatureName],
    logger: Logger,
    operating_mode: OperatingMode,
    parsed_args: Namespace,
    pvt_collector: pvt.PVT,
    *,
    override_ambient_temperature: Optional[float] = None,
    override_collector_input_temperature: Optional[float] = None,
    override_irradiance: Optional[float] = None,
    override_mass_flow_rate: Optional[float] = None,
    override_wind_speed: Optional[float] = None,
    resolution: int = COARSE_RUN_RESOLUTION,
    run_depth: int = 1,
    running_system_temperature_vector: Optional[List[float]] = None,
) -> Tuple[Optional[List[float]], Dict[float, SystemData]]:
    """
    Determines the initial system temperatures for the run.

    This function is called recursively until the temperature vector outputted by the
    system is consistent within some error given by the
    :module:`INITIAL_CONDITION_PRECISION` variable.

    :param number_of_pipes:
        The number of pipes on the base of the hot-water absorber.

    :param layers:
        The layer being used for the run.

    :param logger:
        The logger for the run.

    :param operating_mode:
        The operating mode for the run.

    :param parsed_args:
        The parsed command-line arguments.

    :param pvt_collector:
        The :class:`pvt.PVT` instance representing the pvt panel being modelled.

    :param override_ambient_tempearture:
        If specified, this can be used as a value to override the weather forecaster's
        inbuilt ambient temperature.

    :param override_mass_flow_rate:
        If specified, this can be used as a value to override the collector input file's
        default mass-flow rate.

    :param resolution:
        The resolution for the run, measured in seconds.

    :param run_depth:
        The depth of the recursion.

    :param running_system_temperature_vector:
        The current vector for the system temperatures, used to commence a run and
        compare with the previous run.

    :return:
        A `tuple` containing:
        - the initial system temperature which fits within the desired resolution, if
          applicable, else `None`;
        - the system data from the sun which satisfies the desired consistency.

    """

    # Fetch the initial temperature conditions if not passed in:
    if running_system_temperature_vector is None:
        if operating_mode.coupled:
            running_system_temperature_vector = [
                DEFAULT_SYSTEM_TEMPERATURE
            ] * index_handler.num_temperatures(pvt_collector)
        else:
            running_system_temperature_vector = [DEFAULT_SYSTEM_TEMPERATURE] * (
                index_handler.num_temperatures(pvt_collector) - 3
            )

    # Call the model to generate the output of the run.
    logger.info("Running the model. Run number %s.", run_depth)
    final_temperature_vector, system_data = pvt_system_main(
        parsed_args.average_irradiance,
        parsed_args.cloud_efficacy_factor,
        parsed_args.disable_logging,
        parsed_args.exchanger_data_file,
        parsed_args.initial_month,
        running_system_temperature_vector,
        layers,
        parsed_args.location,
        operating_mode,
        parsed_args.portion_covered,
        parsed_args.pvt_data_file,
        resolution,
        not parsed_args.skip_2d_output,
        parsed_args.tank_data_file,
        parsed_args.use_pvgis,
        parsed_args.verbose,
        parsed_args.x_resolution,
        parsed_args.y_resolution,
        days=parsed_args.days,
        minutes=parsed_args.minutes,
        months=parsed_args.months,
        override_ambient_temperature=override_ambient_temperature,
        override_collector_input_temperature=override_collector_input_temperature,
        override_irradiance=override_irradiance,
        override_mass_flow_rate=override_mass_flow_rate
        if override_mass_flow_rate is not None
        else parsed_args.mass_flow_rate,
        override_wind_speed=override_wind_speed,
        run_number=run_depth,
        start_time=parsed_args.start_time,
    )

    # If in verbose mode, output average, min, and max temperatures.
    if parsed_args.verbose:
        _determine_and_print_average_temperatures(logger, parsed_args, system_data)

    # If all the temperatures are within the desired limit, return the temperatures.
    if all(
        abs(final_temperature_vector - running_system_temperature_vector)
        <= INITIAL_CONDITION_PRECISION
    ):
        logger.info(
            "Initial temperatures consistent. Max difference: %sK",
            max(abs(final_temperature_vector - running_system_temperature_vector)),
        )
        return final_temperature_vector.tolist(), system_data

    if operating_mode.decoupled:
        logger.info("Steady-state run determined at convergence precision.")
        return final_temperature_vector.tolist(), system_data

    logger.info(
        "Initial temperatures not consistent. Max difference: %sK",
        max(abs(final_temperature_vector - running_system_temperature_vector)),
    )
    # Otherwise, call the method recursively.
    return _determine_consistent_conditions(
        number_of_pipes,
        layers,
        logger,
        operating_mode,
        parsed_args,
        pvt_collector,
        override_ambient_temperature=override_ambient_temperature,
        override_collector_input_temperature=override_collector_input_temperature,
        override_irradiance=override_irradiance,
        override_mass_flow_rate=override_mass_flow_rate
        if override_mass_flow_rate is not None
        else parsed_args.mass_flow_rate,
        override_wind_speed=override_wind_speed,
        resolution=resolution,
        run_depth=run_depth + 1,
        running_system_temperature_vector=final_temperature_vector.tolist(),
    )


def _multiprocessing_determine_consistent_conditions(
    steady_state_run: SteadyStateRun,
    *,
    number_of_pipes: int,
    layers: Set[TemperatureName],
    logger: Logger,
    operating_mode: OperatingMode,
    parsed_args: Namespace,
    pvt_collector: pvt.PVT,
) -> Dict[float, SystemData]:
    """
    Wrapper function around `determine_consistent_conditions` to enable multi-processing

    In order for multi-processing to occur, arguments need to be processed in such a way
    that the _determine_consistent_conditions function can be called using a single
    entry from a steady-state data file.

    :param steady_state_entry:
        A data entry from the steady-state data file.

    :param number_of_pipes:
        The number of pipes on the base of the hot-water absorber.

    :param layers:
        The layer being used for the run.

    :param logger:
        The logger for the run.

    :param operating_mode:
        The operating mode for the run.

    :param parsed_args:
        The parsed command-line arguments.

    :param pvt_collector:
        The :class:`pvt.PVT` instance representing the pvt panel being modelled.

    :return:
        The result of the function call to `_determine_consistent_conditions`.

    """

    _, system_data_entry = _determine_consistent_conditions(
        number_of_pipes,
        layers,
        logger,
        operating_mode,
        parsed_args,
        pvt_collector,
        override_ambient_temperature=steady_state_run.ambient_temperature,
        override_collector_input_temperature=steady_state_run.collector_input_temperature
        + ZERO_CELCIUS_OFFSET,
        override_irradiance=steady_state_run.irradiance,
        override_mass_flow_rate=steady_state_run.mass_flow_rate,
        override_wind_speed=steady_state_run.wind_speed,
    )

    return system_data_entry


def _determine_and_print_average_temperatures(
    logger: Logger, parsed_args: Namespace, system_data: Dict[float, SystemData]
) -> None:
    """
    Determines and prints information about the system temperatures.

    The average, minimum, and maximum temperatures of the various components are
    outputted to the console and the logs.

    :param logger:
        The logger used for the run.

    :param parsed_args:
        The parsed command-line arguments.

    :param system_data:
        The system data ouputted by the run.

    """

    def _print_temperature_info(
        average_temperature_map: Dict[str, float],
        logger: Logger,
        maximum_temperature_map: Dict[str, float],
        minimum_temperature_map: Dict[str, float],
        verbose_mode: bool = False,
    ) -> None:
        """
        Prints out the average temperatures to the console and logger.

        :param average_temperature_map:
            A mapping between temperature name, as a `str`, and the average temperature for
            the run.

        :param logger:
            The logger used.

        :param maximum_temperature_run:
            A mapping between temperature name, as a `str`, and the maximum temperature for
            the run.

        :param minimum_temperature_run:
            A mapping between temperature name, as a `str`, and the minimum temperature for
            the run.

        :param verbose_mode:
            If True, then colouring will be done of the output to indicate that it is
            verbose-mode specific.

        """

        logger.info(
            "Average temperatures for the run in degC:\n%s\n%s",
            "|".join(
                [
                    " {}{}".format(
                        key,
                        " "
                        * (
                            max([len(key) for key in average_temperature_map])
                            - len(key)
                            + 1
                        ),
                    )
                    for key in average_temperature_map
                ]
            ),
            "|".join(
                [
                    " {}{}".format(
                        value,
                        " "
                        * (
                            max([len(key) for key in average_temperature_map])
                            - len(str(value))
                            + 1
                        ),
                    )
                    for value in average_temperature_map.values()
                ]
            ),
        )
        print(
            "{}Average temperatures for the run in degC:\n{}\n{}{}".format(
                BColours.OKTEAL if verbose_mode else "",
                "|".join(
                    [
                        " {}{}".format(
                            key,
                            " "
                            * (
                                max([len(key) for key in average_temperature_map])
                                - len(key)
                                + 1
                            ),
                        )
                        for key in average_temperature_map
                    ]
                ),
                "|".join(
                    [
                        " {}{}".format(
                            value,
                            " "
                            * (
                                max([len(key) for key in average_temperature_map])
                                - len(str(value))
                                + 1
                            ),
                        )
                        for value in average_temperature_map.values()
                    ]
                ),
                BColours.ENDC,
            )
        )

        logger.info(
            "Maximum temperatures for the run in degC:\n%s\n%s",
            "|".join(
                [
                    " {}{}".format(
                        key,
                        " "
                        * (
                            max([len(key) for key in maximum_temperature_map])
                            - len(key)
                            + 1
                        ),
                    )
                    for key in maximum_temperature_map
                ]
            ),
            "|".join(
                [
                    " {}{}".format(
                        value,
                        " "
                        * (
                            max([len(key) for key in maximum_temperature_map])
                            - len(str(value))
                            + 1
                        ),
                    )
                    for value in maximum_temperature_map.values()
                ]
            ),
        )
        print(
            "{}Maximum temperatures for the run in degC:\n{}\n{}{}".format(
                BColours.OKTEAL if verbose_mode else "",
                "|".join(
                    [
                        " {}{}".format(
                            key,
                            " "
                            * (
                                max([len(key) for key in maximum_temperature_map])
                                - len(key)
                                + 1
                            ),
                        )
                        for key in maximum_temperature_map
                    ]
                ),
                "|".join(
                    [
                        " {}{}".format(
                            value,
                            " "
                            * (
                                max([len(key) for key in maximum_temperature_map])
                                - len(str(value))
                                + 1
                            ),
                        )
                        for value in maximum_temperature_map.values()
                    ]
                ),
                BColours.ENDC,
            )
        )

        logger.info(
            "Minimum temperatures for the run in degC:\n%s\n%s",
            "|".join(
                [
                    " {}{}".format(
                        key,
                        " "
                        * (
                            max([len(key) for key in minimum_temperature_map])
                            - len(key)
                            + 1
                        ),
                    )
                    for key in minimum_temperature_map
                ]
            ),
            "|".join(
                [
                    " {}{}".format(
                        value,
                        " "
                        * (
                            max([len(key) for key in minimum_temperature_map])
                            - len(str(value))
                            + 1
                        ),
                    )
                    for value in minimum_temperature_map.values()
                ]
            ),
        )
        print(
            "{}Minimum temperatures for the run in degC:\n{}\n{}{}".format(
                BColours.OKTEAL if verbose_mode else "",
                "|".join(
                    [
                        " {}{}".format(
                            key,
                            " "
                            * (
                                max([len(key) for key in minimum_temperature_map])
                                - len(key)
                                + 1
                            ),
                        )
                        for key in minimum_temperature_map
                    ]
                ),
                "|".join(
                    [
                        " {}{}".format(
                            value,
                            " "
                            * (
                                max([len(key) for key in minimum_temperature_map])
                                - len(str(value))
                                + 1
                            ),
                        )
                        for value in minimum_temperature_map.values()
                    ]
                ),
                BColours.ENDC,
            )
        )

    # Determine the average, minimum, and maximum temperatures.
    average_temperature_map = {
        "pv": round(mean({entry.pv_temperature for entry in system_data.values()}), 3),
        "absorber": round(
            mean({entry.absorber_temperature for entry in system_data.values()}), 3
        ),
        "htf": round(
            mean({entry.bulk_water_temperature for entry in system_data.values()}),
            3,
        ),
    }
    maximum_temperature_map = {
        "pv": max({round(entry.pv_temperature, 3) for entry in system_data.values()}),
        "absorber": max(
            {round(entry.absorber_temperature, 3) for entry in system_data.values()}
        ),
        "htf": max(
            {round(entry.bulk_water_temperature, 3) for entry in system_data.values()}
        ),
    }
    minimum_temperature_map = {
        "pv": min({round(entry.pv_temperature, 3) for entry in system_data.values()}),
        "absorber": min(
            {round(entry.absorber_temperature, 3) for entry in system_data.values()}
        ),
        "htf": min(
            {round(entry.bulk_water_temperature, 3) for entry in system_data.values()}
        ),
    }

    if "dg" in parsed_args.layers:
        average_temperature_map["upper_glass"] = round(
            mean({entry.upper_glass_temperature for entry in system_data.values()}), 3  # type: ignore
        )
        maximum_temperature_map["upper_glass"] = max(
            {round(entry.upper_glass_temperature, 3) for entry in system_data.values()}  # type: ignore
        )
        minimum_temperature_map["upper_glass"] = min(
            {round(entry.upper_glass_temperature, 3) for entry in system_data.values()}  # type: ignore
        )

    if "g" in parsed_args.layers:
        average_temperature_map["glass"] = round(
            mean({entry.glass_temperature for entry in system_data.values()}), 3  # type: ignore
        )
        maximum_temperature_map["glass"] = max(
            {round(entry.glass_temperature, 3) for entry in system_data.values()}  # type: ignore
        )
        minimum_temperature_map["glass"] = min(
            {round(entry.glass_temperature, 3) for entry in system_data.values()}  # type: ignore
        )

    if not parsed_args.decoupled:
        average_temperature_map["tank"] = round(
            mean({entry.tank_temperature for entry in system_data.values()}), 3  # type: ignore
        )
        maximum_temperature_map["tank"] = max(
            {round(entry.tank_temperature, 3) for entry in system_data.values()}  # type: ignore
        )
        minimum_temperature_map["tank"] = min(
            {round(entry.tank_temperature, 3) for entry in system_data.values()}  # type: ignore
        )

    # Print these out to the console.
    logger.info("Average, minimum, and maximum temperatures determined.")
    _print_temperature_info(
        average_temperature_map,
        logger,
        maximum_temperature_map,
        minimum_temperature_map,
        True,
    )


def main(args) -> None:  # pylint: disable=too-many-branches
    """
    The main module for the code.

    :param args:
        The command-line arguments passed into the component.

    """

    # Print the header string.
    print(PVT_HEADER_STRING)

    # Parse the arguments passed in.
    parsed_args = argparser.parse_args(args)

    # Initialise logging.
    logger = get_logger(parsed_args.disable_logging, LOGGER_NAME, parsed_args.verbose)
    logger.info(
        "%s PVT model instantiated. %s\nCommand: %s", "=" * 20, "=" * 20, " ".join(args)
    )
    print(
        "HEAT model of a hybrid PV-T collector instantiated{}.".format(
            f"{BColours.OKTEAL} in verbose mode{BColours.ENDC}"
            if parsed_args.verbose
            else ""
        )
    )

    print(
        "Verifying input information and arguments {}    ".format(
            "." * 21,
        ),
        end="",
    )
    try:
        # Validate the CLI arguments.
        layers = argparser.check_args(
            parsed_args,
            logger,
            read_yaml(parsed_args.pvt_data_file)["absorber"]["number_of_pipes"],
        )
    except argparser.ArgumentMismatchError:
        print(FAILED)
        raise
    print(DONE)

    # Parse the PVT system information and generate a PVT panel based on the args for
    # use in Fourier-number calculations.
    pvt_collector = pvt_collector_from_path(
        layers,
        logger,
        parsed_args.mass_flow_rate,
        parsed_args.portion_covered,
        parsed_args.pvt_data_file,
        parsed_args.x_resolution,
        parsed_args.y_resolution,
    )
    logger.debug(
        "PV-T panel elements:\n  %s",
        "\n  ".join(
            [
                f"{element_coordinates}: {element}"
                for element_coordinates, element in pvt_collector.elements.items()
            ]
        ),
    )

    if os.path.isfile(f"{parsed_args.output}.yaml"):
        logger.info("The output YAML file specified already exists. Moving...")
        os.rename(f"{parsed_args.output}.yaml", f"{parsed_args.output}.yaml.1")
        logger.info("Output file successfully moved.")
    if os.path.isfile(f"{parsed_args.output}.json"):
        logger.info("The output YAML file specified already exists. Moving...")
        os.rename(f"{parsed_args.output}.json", f"{parsed_args.output}.json.1")
        logger.info("Output file successfully moved.")

    # Instantiate a hot-water tank instance based on the data.
    if parsed_args.decoupled:
        hot_water_tank: Optional[tank.Tank] = None
    else:
        hot_water_tank = hot_water_tank_from_path(parsed_args.tank_data_file)

    logger.debug(
        "PVT system information successfully parsed:\n%s\n%s",
        str(pvt_collector),
        str(hot_water_tank) if hot_water_tank is not None else "NO HOT-WATER TANK",
    )

    # Determine the Fourier number for the PV-T panel.
    logger.info("Beginning Fourier number calculation.")
    _calculate_and_print_fourier_numbers(
        hot_water_tank, logger, parsed_args, pvt_collector
    )

    # Determine the operating mode of the system.
    operating_mode = OperatingMode(not parsed_args.decoupled, parsed_args.dynamic)

    if operating_mode.dynamic:
        logger.info(
            "Running a dynamic and %scoupled system.",
            "de" if operating_mode.decoupled else "",
        )
        print(
            "{}Running a dynamic and {}coupled system.{}".format(
                BColours.OKGREEN,
                "de" if operating_mode.decoupled else "",
                BColours.ENDC,
            )
        )

        if operating_mode.coupled:
            # The initial system conditions for the run need to be determined so that
            # they match up at the start and end of the time period being modelled,
            # i.e., so that there is no transient in the data.

            # Iterate to determine the initial conditions for the run at a rough
            # resolution.
            logger.info(
                "Determining consistent initial conditions at coarse resolution."
            )
            print(
                "Determining consistent initial conditions via successive runs at "
                f"{COARSE_RUN_RESOLUTION}s resolution."
            )

            initial_system_temperature_vector, _ = _determine_consistent_conditions(
                pvt_collector.absorber.number_of_pipes,
                layers,
                logger,
                operating_mode,
                parsed_args,
                pvt_collector,
            )
            logger.info(
                "Consistent initial conditions determined at coarse resolution."
            )

            # Run again at a fine resolution to better determine the initial conditions.
            print(
                "Rough initial conditions determined at coarse resolution, refining via "
                f"successive runs at CLI resolution of {parsed_args.resolution}s."
            )
            initial_system_temperature_vector, _ = _determine_consistent_conditions(
                pvt_collector.absorber.number_of_pipes,
                layers,
                logger,
                operating_mode,
                parsed_args,
                pvt_collector,
                resolution=parsed_args.resolution,
                running_system_temperature_vector=initial_system_temperature_vector,
            )
            if initial_system_temperature_vector is None:
                raise ProgrammerJudgementFault(
                    "{}The initial system conditions were not returned as ".format(
                        BColours.FAIL
                    )
                    + "expected when performing a coupled run.{}".format(BColours.ENDC)
                )
            logger.info(
                "Initial system temperatures successfully determined to %sK precision.",
                INITIAL_CONDITION_PRECISION,
            )
            print(
                f"Initial conditions determined to {INITIAL_CONDITION_PRECISION}K precision."
            )

        else:
            # In the case of a decoupled and dynamic model, usually a step change is
            # being modelled. In these instances, the system is run for some time,
            # determined by the data fed in, and the initial conditions can simply be
            # taken as the default conditions.
            initial_system_temperature_vector = [DEFAULT_SYSTEM_TEMPERATURE] * (
                index_handler.num_temperatures(pvt_collector) - 3
            )

        logger.info(
            "Running the model at the CLI resolution of %ss.", parsed_args.resolution
        )
        print(
            f"Running the model at the high CLI resolution of {parsed_args.resolution}s."
        )
        # Run the model at this higher resolution.
        _, system_data = pvt_system_main(
            parsed_args.average_irradiance,
            parsed_args.cloud_efficacy_factor,
            parsed_args.disable_logging,
            parsed_args.exchanger_data_file,
            parsed_args.initial_month,
            initial_system_temperature_vector,
            layers,
            parsed_args.location,
            operating_mode,
            parsed_args.portion_covered,
            parsed_args.pvt_data_file,
            parsed_args.resolution,
            not parsed_args.skip_2d_output,
            parsed_args.tank_data_file,
            parsed_args.use_pvgis,
            parsed_args.verbose,
            parsed_args.x_resolution,
            parsed_args.y_resolution,
            days=parsed_args.days,
            minutes=parsed_args.minutes,
            months=parsed_args.months,
            override_ambient_temperature=parsed_args.ambient_temperature,
            override_collector_input_temperature=parsed_args.collector_input_temperature,
            override_irradiance=parsed_args.solar_irradiance,
            override_mass_flow_rate=parsed_args.mass_flow_rate,
            override_wind_speed=parsed_args.wind_speed,
            run_number=1,
            start_time=parsed_args.start_time,
        )

    elif operating_mode.decoupled and operating_mode.steady_state:
        logger.info("Running a steady-state and decoupled system.")
        print(
            f"{BColours.OKGREEN}"
            "Running a steady-state and decoupled system."
            f"{BColours.ENDC}"
        )
        # Set up a holder for information about the system.
        system_data = dict()

        if parsed_args.steady_state_data_file is not None:
            # If specified, parse the steady-state data file.
            steady_state_runs = [
                SteadyStateRun.from_data(entry)
                for entry in read_yaml(parsed_args.steady_state_data_file)
            ]

            for steady_state_run in steady_state_runs:
                if parsed_args.ambient_temperature is not None:
                    steady_state_run.ambient_temperature = (
                        parsed_args.ambient_temperature
                    )
                if parsed_args.collector_input_temperature is not None:
                    steady_state_run.collector_input_temperature = (
                        parsed_args.collector_input_temperature
                    )
                if parsed_args.solar_irradiance is not None:
                    steady_state_run.irradiance = parsed_args.solar_irradiance
                if parsed_args.wind_speed is not None:
                    steady_state_run.wind_speed = parsed_args.wind_speed

            logger.info(
                "%s run%s will be attempted based on the input data file.",
                len(steady_state_runs),
                "s" if len(steady_state_runs) > 1 else "",
            )
            print(
                f"{BColours.OKGREEN}"
                + "{} run{} will be attempted based on the ".format(
                    len(steady_state_runs), "s" if len(steady_state_runs) > 1 else ""
                )
                + f"information in {parsed_args.steady_state_data_file}."
                + f"{BColours.ENDC}"
            )

            # with Pool(8) as worker_pool:
            #     logger.info(
            #         "A multi-process worker pool will be used to parallelise the task."
            #     )
            #     multi_processing_output = worker_pool.map(
            #         partial(
            #             _multiprocessing_determine_consistent_conditions,
            #             number_of_pipes=pvt_collector.absorber.number_of_pipes,
            #             layers=layers,
            #             logger=logger,
            #             operating_mode=operating_mode,
            #             parsed_args=parsed_args,
            #             pvt_collector=pvt_collector,
            #         ),
            #         steady_state_runs
            #     )
            #     logger.info("Multi-process worker pool successfully completed.")

            for run_number, steady_state_run in enumerate(
                tqdm(steady_state_runs, desc="steady state runs", unit="run"), 1
            ):
                logger.info(
                    "Carrying out steady-state run %s of %s.",
                    run_number,
                    len(steady_state_runs),
                )
                try:
                    output = _multiprocessing_determine_consistent_conditions(
                        steady_state_run,
                        number_of_pipes=pvt_collector.absorber.number_of_pipes,
                        layers=layers,
                        logger=logger,
                        operating_mode=operating_mode,
                        parsed_args=parsed_args,
                        pvt_collector=pvt_collector,
                    )
                except RecursionError as e:
                    logger.error(
                        "Recursion error processing steady state run.\nRun: %s\nMsg: %s",
                        steady_state_run,
                        str(e),
                    )
                    continue
                except DivergentSolutionError as e:
                    logger.error(
                        "A divergent solution occurred - have you considered the "
                        "difference between Celcius and Kelvin in all your units, "
                        "especially override CLI units. Consider checking this before "
                        "investigating further.\nRun attempted: %s\nError: %s",
                        steady_state_run,
                        str(e)
                    )
                    continue

                for key, value in output.items():
                    system_data[f"run_{run_number}_T_in_{key}degC"] = value

        else:
            raise ProgrammerJudgementFault(
                "Steady-state data must be specified with a JSON data file."
            )
    else:
        raise ProgrammerJudgementFault(
            "The system needs to either be run in steady-state or dynamic modes, with "
            "either `--steady-state` or `--dynamic` specified on the CLI."
        )

    # Save the data ouputted by the model.
    logger.info("Saving output data to: %s.json.", parsed_args.output)
    save_data(FileType.JSON, logger, operating_mode, parsed_args.output, system_data)
    print(f"Model output successfully saved to {parsed_args.output}.json.")

    # If in verbose mode, output average, min, and max temperatures.
    if parsed_args.verbose:
        _determine_and_print_average_temperatures(logger, parsed_args, system_data)

    if parsed_args.skip_analysis:
        logger.info("Analysis will be skippted.")
        print("Skipping analysis. This can be run manually later.")
    else:
        # Conduct analysis of the data.
        logger.info("Conducting analysis.")
        print("Conducting analysis.")
        analysis.analyse(f"{parsed_args.output}.json")  # type: ignore
        print("Analysis complete. Figures can be found in `./figures`.")
        logger.info("Analysis complete.")

    logger.info("Exiting.")


if __name__ == "__main__":
    try:
        main(sys.argv[1:])
    except DivergentSolutionError:
        print(
            "A divergent solution occurred - have you considered the difference "
            "between Celcius and Kelvin in all your units, especially override CLI "
            "units. Consider checking this before investigating further."
        )
    except Exception:
        print("An exception occured. See /logs for details.")
        raise



#!/usr/bin/python3.7
########################################################################################
# pvt_collector/absorber.py - Represents a absorber within a PVT panel.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The absorber module for the PV-T model.

This module represents a thermal absorber within a PV-T panel.

"""

import logging
import math

from ..__utils__ import (
    CollectorParameters,
    PVT_SYSTEM_MODEL_LOGGER_NAME,
    OpticalLayerParameters,
)
from .__utils__ import (
    OpticalLayer,
)

__all__ = ("Collector",)

# Get the logger for the run.
logger = logging.getLogger(PVT_SYSTEM_MODEL_LOGGER_NAME)


class Collector(OpticalLayer):
    """
    Represents the thermal absorber (lower) layer of the PV-T panel.

    .. attribute:: htf_heat_capacity
        The heat capacity of the heat-transfer fluid passing through the absorber,
        measured in Joules per kilogram Kelvin.

    .. attribute:: mass_flow_rate
        The mass flow rate of heat-transfer fluid through the absorber, measured in
        kilograms per second.

    .. attribute:: output_water_temperature
        The temperature of the water outputted by the layer, measured in Kelvin.

    .. attribute:: pump_power
        The power consumed by the water pump, measured in Watts.

    """

    # Pirvate Attributes:
    #
    # .. attribute:: _mass_flow_rate
    #   The mass flow rate of heat-trasnfer fluid through the absorber, measured in
    #   Litres per hour.
    #

    def __init__(self, absorber_params: CollectorParameters) -> None:
        """
        Instantiate a absorber layer.

        :param absorber_params:
            The parameters needed to instantiate the absorber.

        """

        super().__init__(
            OpticalLayerParameters(
                absorber_params.conductivity,
                absorber_params.density,
                absorber_params.heat_capacity,
                absorber_params.thickness,
                absorber_params.transmissivity,
                absorber_params.absorptivity,
                absorber_params.emissivity,
            )
        )

        self.htf_heat_capacity = absorber_params.htf_heat_capacity
        self.inner_pipe_diameter = absorber_params.inner_pipe_diameter
        self.length = absorber_params.length
        self._mass_flow_rate = absorber_params.mass_flow_rate
        self.number_of_pipes = absorber_params.number_of_pipes
        self.outer_pipe_diameter = absorber_params.outer_pipe_diameter
        self.pipe_density = absorber_params.pipe_density

    def __repr__(self) -> str:
        """
        Returns a nice representation of the layer.

        :return:
            A `str` giving a nice representation of the layer.

        """

        return (
            "Collector("
            f"absorptivity: {self.absorptivity}, "
            f"conductivity: {self.conductivity}W/m^2*K, "
            f"desntiy: {self.density}kg/m^3, "
            f"emissivity: {self.emissivity}, "
            f"heat_capacity: {self.heat_capacity}J/kg*K, "
            f"htf_heat_capacity: {self.htf_heat_capacity}J/kg*K, "
            f"inner_pipe_diameter: {self.inner_pipe_diameter}m, "
            f"length: {self.length}m, "
            f"mass_flow_rate: {self.mass_flow_rate}kg/s, "
            f"outer_pipe_diameter: {self.outer_pipe_diameter}m, "
            f"thickness: {self.thickness}m, "
            f"transmissivity: {self.transmissivity}"
            ")"
        )

    @property
    def htf_surface_area(self) -> float:
        """
        Returns the contact area between the HTF and the absorber, measured in m^2.

        :return:
            The contact surface area, between the absorber (i.e., the pipes) and the
            HTF passing through the pipes.
            A single pass is assumed, with multiple pipes increasing the area, rather
            than the length, of the absorber.

        """

        return (
            self.number_of_pipes  # [pipes]
            * math.pi
            * self.inner_pipe_diameter  # [m]
            * self.length  # [m]
        )

    @property
    def htf_volume(self) -> float:
        """
        Returns the volume of HTF that can be held within the absorber, measured in m^3

        :return:
            The volume of the HTF within the absorber, measured in meters cubed.

        """

        return (
            self.number_of_pipes  # [pipes]
            * math.pi
            * (self.inner_pipe_diameter / 2) ** 2  # [m^2]
            * self.length
        )

    @property
    def mass_flow_rate(self) -> float:
        """
        Return the mass-flow rate in kilograms per second.

        :return:
            d/dt(M) in kg/s

        """

        return self._mass_flow_rate / (3600)  # [kg/s]



#!/usr/bin/python3.7
########################################################################################
# physics_utils.py - Utility module containing Physics equations.
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
The physics utility module for the PVT model's PVT panel component.

"""

import math

from typing import List, Union

from numpy import ndarray

from .. import physics_utils
from . import pvt

from ...__utils__ import BColours, ProgrammerJudgementFault
from ..__utils__ import WeatherConditions

__all__ = (
    "glass_glass_air_gap_resistance",
    "glass_pv_air_gap_resistance",
    "insulation_thermal_resistance",
)


def _conductive_heat_transfer_coefficient_with_gap(
    air_gap_thickness: float,
    average_surface_temperature: float,
    pvt_collector: pvt.PVT,
    weather_conditions: WeatherConditions,
) -> float:
    """
    Computes the conductive heat transfer between the two layers, measured in W/m^2*K.

    The value computed is positive if the heat transfer is from the source to the
    destination, as determined by the arguments, and negative if the flow of heat is
    the reverse of what is implied via the parameters.

    The value for the heat transfer is returned in Watts.

    :param air_gap_thickness:
        The thickness of the air gap between the PV and glass layers.

    :param average_surface_temperature:
        The average temperature of the surfaces across which the heat transfer is taking
        place.

    :param pvt_collector:
        The PVT panel being modelled.

    :param weather_conditions:
        The weather conditions at the time step being investigated.

    :return:
        The heat transfer coefficient, in Watts per meter squared Kelvin, between the
        two layers.

    """

    if pvt_collector.air_gap_thickness is None:
        raise ProgrammerJudgementFault(
            "{}Conductive heat transfer requested with no air gap.{}".format(
                BColours.FAIL, BColours.ENDC
            )
        )
    air_gap_rayleigh_number = physics_utils.rayleigh_number(
        pvt_collector.air_gap_thickness,
        average_surface_temperature,
        pvt_collector.tilt_in_radians,
        weather_conditions,
    )

    first_corrective_term: float = 1 - 1708 / air_gap_rayleigh_number
    first_corrective_term = max(first_corrective_term, 0)

    second_corrective_term: float = 1 - (
        1708 * (math.sin(1.8 * pvt_collector.tilt_in_radians)) ** 1.6
    ) / (air_gap_rayleigh_number * math.cos(pvt_collector.tilt_in_radians))
    second_corrective_term = max(second_corrective_term, 0)

    third_corrective_term: float = (
        (air_gap_rayleigh_number * math.cos(pvt_collector.tilt_in_radians)) / 5830
    ) ** 0.33 - 1
    third_corrective_term = max(third_corrective_term, 0)

    return (
        weather_conditions.thermal_conductivity_of_air  # [W/m*K]
        / air_gap_thickness  # [m]
    ) * (
        1
        + 1.44 * first_corrective_term * second_corrective_term
        + third_corrective_term
    )


def glass_absorber_air_gap_resistance(
    pvt_collector: pvt.PVT,
    surface_temperature: float,
    weather_conditions: WeatherConditions,
) -> float:
    """
    Returns the thermal resistance of the air gap between the absorber and glass layers.

    :param pvt_collector:
        The :class:`pvt.PVT` instance representing the pvt panel being modelled.

    :param surface_temperature:
        The average temperature of the two surfaces either side of the air gap.

    :param weather_conditions:
        The weather conditions at the time step being investigated.

    :return:
        The thermal resistance, measured in Kelvin meter squared per Watt.

    """

    if pvt_collector.glass is None or pvt_collector.air_gap_thickness is None:
        raise ProgrammerJudgementFault(
            "{}Resistance across an air gap could not be computed due to no ".format(
                BColours.FAIL
            )
            + "glass layer being present.{}".format(BColours.ENDC)
        )

    return (
        pvt_collector.eva.thickness / pvt_collector.eva.conductivity
        + pvt_collector.glass.thickness / pvt_collector.glass.conductivity
        + pvt_collector.absorber.thickness / (2 * pvt_collector.absorber.conductivity)
        + pvt_collector.glass.thickness / (2 * pvt_collector.glass.conductivity)
        + 1
        / _conductive_heat_transfer_coefficient_with_gap(
            pvt_collector.air_gap_thickness,
            surface_temperature,
            pvt_collector,
            weather_conditions,
        )
    )


def glass_glass_air_gap_resistance(
    pvt_collector: pvt.PVT,
    surface_temperature: float,
    weather_conditions: WeatherConditions,
) -> float:
    """
    Returns the thermal resistance of the air gap between two glass layers.

    :param pvt_collector:
        The :class:`pvt.PVT` instance representing the pvt panel being modelled.

    :param surface_temperature:
        The average temperature of the two surfaces either side of the air gap.

    :param weather_conditions:
        The weather conditions at the time step being investigated.

    :return:
        The thermal resistance, measured in Kelvin meter squared per Watt.

    """

    if (
        pvt_collector.glass is None
        or pvt_collector.air_gap_thickness is None
        or pvt_collector.upper_glass is None
    ):
        raise ProgrammerJudgementFault(
            "{}Resistance across an air gap could not be computed due to no ".format(
                BColours.FAIL
            )
            + "glass layer being present.{}".format(BColours.ENDC)
        )

    return (
        pvt_collector.eva.thickness / pvt_collector.eva.conductivity
        + pvt_collector.upper_glass.thickness / pvt_collector.upper_glass.conductivity
        + pvt_collector.glass.thickness / (2 * pvt_collector.glass.conductivity)
        + pvt_collector.upper_glass.thickness
        / (2 * pvt_collector.upper_glass.conductivity)
        + 1
        / _conductive_heat_transfer_coefficient_with_gap(
            pvt_collector.air_gap_thickness,
            surface_temperature,
            pvt_collector,
            weather_conditions,
        )
    )


def glass_pv_air_gap_resistance(
    pvt_collector: pvt.PVT,
    surface_temperature: float,
    weather_conditions: WeatherConditions,
) -> float:
    """
    Returns the thermal resistance of the air gap between the PV and glass layers.

    :param pvt_collector:
        The :class:`pvt.PVT` instance representing the pvt panel being modelled.

    :param surface_temperature:
        The average temperature of the two surfaces either side of the air gap.

    :param weather_conditions:
        The weather conditions at the time step being investigated.

    :return:
        The thermal resistance, measured in Kelvin meter squared per Watt.

    """

    if pvt_collector.glass is None or pvt_collector.air_gap_thickness is None:
        raise ProgrammerJudgementFault(
            "{}Resistance across an air gap could not be computed due to no ".format(
                BColours.FAIL
            )
            + "glass layer being present.{}".format(BColours.ENDC)
        )

    return (
        pvt_collector.eva.thickness / pvt_collector.eva.conductivity
        + pvt_collector.glass.thickness / pvt_collector.glass.conductivity
        + pvt_collector.pv.thickness / (2 * pvt_collector.pv.conductivity)
        + pvt_collector.glass.thickness / (2 * pvt_collector.glass.conductivity)
        + 1
        / _conductive_heat_transfer_coefficient_with_gap(
            pvt_collector.air_gap_thickness,
            surface_temperature,
            pvt_collector,
            weather_conditions,
        )
    )


def insulation_thermal_resistance(
    best_guess_temperature_vector: Union[List[float], ndarray],
    pvt_collector: pvt.PVT,
    source_index: int,
    weather_conditions: WeatherConditions,
) -> float:
    """
    Returns the thermal resistance between the back layer of the absorber and air.

    Insulation on the back of the PV-T absorber causes there to be some thermal
    resistance to the heat transfer out of the back of the thermal absorber. This
    value is computed here and returned.

    :param best_guess_temperature_vector:
        The best-guess at the temperature vector for the system at the current time
        step.

    :param pvt_collector:
        The :class:`pvt.PVT` instance representing the pvt panel being modelled.

    :param source_index:
        The index of the element emitting to the environment.

    :param weather_conditions:
        The weather conditions at the current time step.

    :return:
        The thermal resistance between the back layer of the absorber and the
        surrounding air, measured in meter squared Kelvin per Watt.

    """

    return (
        pvt_collector.insulation.thickness / pvt_collector.insulation.conductivity
        + 1
        / physics_utils.free_heat_transfer_coefficient_of_air(
            pvt_collector,
            best_guess_temperature_vector[source_index],
            weather_conditions,
        )
    )



#!/usr/bin/python3.7
########################################################################################
# enforcement/pylint_enforcement.py
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2021
########################################################################################

"""
Script for ensuring that all pylint errors are declared correctly in the pylint yaml file.

All uses of the `pylint: disable` pylint declaration must be declared in pylint_enforcement.yaml
in the enforcement folder.

"""

import os
import subprocess

from dataclasses import dataclass

import yaml

# String required in all enforcement entries.
JUSTIFICATION_STRING = "justification"
# Path to the YAML enforcement file.
YAML_ENFORCEMENT_FILE = os.path.join("enforcement", "pylint_enforcement.yaml")


class EnforcementError(Exception):
    """
    Raised when an enforcement error occurs.

    """

    def __init__(self, msg: str) -> None:
        """
        Instantiate a :class:`EnforcementError` instance.

        """

        super().__init__(
            f"Not all pylint 'pylint: disable' uses were declared in the yaml file: {msg}"
        )


@dataclass
class PylintDisable:
    """
    Represents a single use of the "pylint: disable" declaration.

    .. attribute:: filename
        The name of the file containing the usage.

    .. attribute:: flag
        The flag disabled.

    .. attribute:: line
        The line on which the flag is used.

    """

    filename: str
    flag: str
    line: int

    def __eq__(self, other) -> bool:
        """
        Determines if two instances are equal.

        """

        return str(self) == str(other)

    def __lt__(self, other) -> bool:
        """
        Determines if the instance is less than the other.

        """

        return str(self) < str(other)

    def __repr__(self) -> str:
        """
        Representation of the type ignore usage.

        """

        return f"{self.filename}:{self.line}:{self.flag}"

    def __str__(self) -> str:
        """
        Representation of the type ignore usage.

        """

        return f"{self.filename}:{self.line}:{self.flag}"


def main() -> None:
    """
    Main method for the pylint enforcement module.

    """

    # Load the declared uses of "pylint: disable".
    with open(YAML_ENFORCEMENT_FILE) as f:
        pylint_disable_declarations = yaml.safe_load(f)

    processed_pylint_disable_declarations = sorted(
        [
            PylintDisable(entry["file"], entry["flag"], entry["line"])
            for entry in pylint_disable_declarations
        ]
    )

    if not all(
        (JUSTIFICATION_STRING in entry for entry in pylint_disable_declarations)
    ):
        raise EnforcementError(
            "Not all entries were justified: {}".format(
                "; ".join(
                    [
                        entry
                        for entry in pylint_disable_declarations
                        if JUSTIFICATION_STRING not in entry
                    ]
                )
            )
        )

    # Run from the above directory if needed.
    if "pvt_model" not in os.listdir():
        directory_prefix: str = ".." + os.path.sep
    else:
        directory_prefix = ""

    # Determine the actual uses
    pylint_disable_uses = (
        subprocess.run(
            f"grep 'pylint: disable' {directory_prefix}pvt_model -rn",
            capture_output=True,
            check=True,
            shell=True,
        )
        .stdout.decode("utf-8")
        .split("\n")
    )
    pylint_disable_uses = [
        entry.replace("# pylint: disable", "") for entry in pylint_disable_uses
    ]
    pylint_disable_uses = [entry for entry in pylint_disable_uses if entry != ""]

    # Process this into usable data.
    processed_pylint_disable_uses = sorted(
        [
            PylintDisable(
                entry.split(":")[0],
                ":".join(entry.split(":")[2:]).split("=")[1],
                int(entry.split(":")[1]),
            )
            for entry in pylint_disable_uses
        ]
    )

    if processed_pylint_disable_declarations != processed_pylint_disable_uses:
        raise EnforcementError(
            "Missing declarations:\n{}\nExcess declarations:\n{}".format(
                "\n".join(
                    [
                        str(entry)
                        for entry in processed_pylint_disable_uses
                        if entry not in processed_pylint_disable_declarations
                    ]
                ),
                "\n".join(
                    [
                        str(entry)
                        for entry in processed_pylint_disable_declarations
                        if entry not in processed_pylint_disable_uses
                    ]
                ),
            )
        )


if __name__ == "__main__":
    main()



#!/usr/bin/python3.7
# type: ignore
########################################################################################
# __utils__.py - The utility module for this, my first, PV-T model! :O
#
# Author: Ben Winchester
# Copyright: Ben Winchester, 2020
########################################################################################

"""
The utility module for the analysis component.

"""

import enum
import os

from logging import Logger
from typing import Any, List, Dict, Optional, Tuple

import json
import re

import numpy

from matplotlib.axes import Axes
from matplotlib import cm
from matplotlib import pyplot as plt
from mpl_toolkits.mplot3d import Axes3D as plt3D

from tqdm import tqdm

try:
    from ..pvt_system.constants import (  # pylint: disable=unused-import
        HEAT_CAPACITY_OF_WATER,
    )
except ModuleNotFoundError:
    import logging

    logging.error(
        "Incorrect module import. Try running with `python3.7 -m pvt_model.analysis`"
    )
    raise


__all__ = (
    "GraphDetail",
    "load_model_data",
    "plot",
    "plot_figure",
    "plot_two_dimensional_figure",
)

# Used to distinguish copled data sets.
COUPLED_DATA_TYPE = "coupled"

# Used to distinguished decoupled data sets.
DECOUPLED_DATA_TYPE = "decoupled"

# Used to distinguish dynamic data sets.
DYNAMIC_DATA_TYPE = "dynamic"

# The directory into which which should be saved
NEW_FIGURES_DIRECTORY: str = "figures"

# Used to identify the "time" data set.
TIME_KEY = "time"

# The directory in which old figures are saved and stored for long-term access
OLD_FIGURES_DIRECTORY: str = "old_figures"

# Used to distinguish steady-state data sets.
STEADY_STATE_DATA_TYPE = "steady_state"

# Name of the steady-state data file.
STEADY_STATE_DATA_FILE_NAME = "autotherm.yaml"

# How many values there should be between each tick on the x-axis
# X_TICK_SEPARATION: int = int(8 * GRAPH_DETAIL.value / 48)
X_TICK_SEPARATION: int = 8

# Which days of data to include
DAYS_TO_INCLUDE: List[bool] = [False, True]

# The resolution to use for the trendline.
TRENDLINE_RESOLUTION = 200


class GraphDetail(enum.Enum):
    """
    The level of detail to go into when graphing.

    .. attribute:: highest
        The highest level of detail - all data points are plotted.

    .. attribute:: high
        A "high" level of detail, to be determined by the analysis script.

    .. attribute:; medium
        A "medium" level of detail, to be determined by the analysis script.

    .. attribute:: low
        A "low" level of detail, to be determined by the analysis script.

    .. attribute:: lowest
        The lowest level of detail, with points only every half an hour.

    """

    highest = 0
    high = 2880
    medium = 720
    low = 144
    lowest = 48


def _annotate_maximum(
    model_data: Dict[Any, Any], y_axis_labels: List[str], axis
) -> None:
    """
    Annotates the maximum value on a plot.

    .. citation:: Taken with permission from:
    https://stackoverflow.com/questions/43374920/how-to-automatically-annotate-maximum-value-in-pyplot/43375405

    :param data:
        The model data to find the maxima from.

    :param y_axis_labels:
        Labels for the y axis to process.

    :param axis:
        The axis object, if relevant.

    """

    # * For each series, determine the maximum data points:
    box_text = ""
    x_series = [data_entry[TIME_KEY] for data_entry in model_data.values()]
    for y_lab in y_axis_labels:
        y_series = [data_entry[y_lab] for data_entry in model_data.values()]
        x_max = x_series[max(enumerate(y_series))[0]]
        y_max = max(y_series)
        box_text += f"max({y_lab})={y_max:.2f} at {x_max}\n"
    box_text.strip()

    # Fetch the axis if necessary.
    if not axis:
        axis = plt.gca()
    bbox_props = dict(boxstyle="square,pad=0.3", fc="w", ec="k", lw=0.72)
    arrowprops = dict(arrowstyle="->", connectionstyle="angle,angleA=0,angleB=85")
    kwargs = dict(
        xycoords="data",
        textcoords="axes fraction",
        arrowprops=arrowprops,
        bbox=bbox_props,
        ha="right",
        va="top",
    )
    axis.ticklabel_format(useOffset=False)
    axis.annotate(box_text, xy=(x_max, y_max), xytext=(0.8, 0.8), **kwargs)


def load_model_data(filename: str) -> Dict[Any, Any]:
    """
    Loads some model_data that was generated by the model.

    :param filename:
        The name of the model_data file to open.

    :return:
        The JSON model_data, loaded as a `dict`.

    """

    with open(filename, "r") as f:
        json_data: Dict[Any, Any] = json.load(f)

    return json_data


def plot(  # pylint: disable=too-many-branches
    label: str,
    use_data_keys: bool,
    x_axis_key: str,
    x_label: str,
    y_label: str,
    model_data: Dict[Any, Any],
    *,
    disable_lines: bool,
    axes=None,
    bar_plot: bool = False,
    colour: str = None,
    hold: bool = False,
    shape: str = "x",
    trendline: bool = False,
) -> Optional[Any]:
    """
    Plots some model_data based on input parameters.

    :param label:
        The name of the model_data label to plot. For example, this could be
        "pv_temperature", in which case, the pv temperature would be plotted at the time
        steps specified in the model.

    :param use_data_keys:
        If specified, the keys of the data will be used. Otherwise, the value provided
        via the x-axis key will be used.

    :param x_axis_key:
        Key used for the x-axis data set. This defaults to the time of day.

    :param x_label:
        The x-axis label.

    :param y_label:
        The label to assign to the y axis when plotting.

    :param resolution:
        The resolution of the model that was run. This is measured in minutes.

    :param model_data:
        The model_data, loaded from the model's output file.

    :param disable_lines:
        If specified, lines will be disabled from the output.

    :param axes:
        If provided, a separate axis is used for plotting the model_data.

    :param bar_plot:
        Whether to plot a line graph (False) or a bar_plot plot (True).

    :param colour:
        The colour to use for the plot.

    :param hold:
        Whether to hold the screen between plots (True) or reset it (False).

    :param shape:
        This sets the shape of the marker for `matplotlib.pyplot` to use when plotting
        the model_data.

    :param trendline:
        Whether to plot a trendline (True) or not (False).

    """

    # Use the data keys themselves if appropriate, otherwise use the x-label provided.
    if use_data_keys:
        x_model_data = [float(entry) for entry in model_data.keys()]
    else:
        x_model_data = [entry[x_axis_key] for entry in model_data.values()]
    y_model_data = [entry[label] for entry in model_data.values()]

    # Sort the data series based on increasing x values if the data is not time.
    if x_axis_key != TIME_KEY:
        x_model_data, y_model_data = (
            list(entry) for entry in zip(*sorted(zip(x_model_data, y_model_data)))
        )

    # If we are not holding the graph, then clear the model_data.
    if not hold:
        plt.clf()
        plt.close("all")

    # Reduce the values on the x axis to be times.
    # x_model_data = [float(item) / (resolution / 60) for item in x_model_data]

    # If needed, compute a trendline.
    if trendline:
        trend = numpy.polyfit(
            x_model_data,
            y_model_data,
            2,
        )
        trendline_x_data = numpy.linspace(
            min(x_model_data), max(x_model_data), TRENDLINE_RESOLUTION
        )
        trendline_y_data = [
            trend[0] * entry ** 2 + trend[1] * entry + trend[2]
            for entry in trendline_x_data
        ]

    if bar_plot:
        if axes is None:
            if colour is None:
                plt.bar(x_model_data, y_model_data, label=label)
            else:
                plt.bar(x_model_data, y_model_data, label=label, color=colour)

        #  otherwise, the model_data needs to be plotted on just on axis.
        else:
            if colour is None:
                line = axes.bar(x_model_data, y_model_data, label=label)
            else:
                line = axes.bar(x_model_data, y_model_data, label=label, color=colour)

    else:
        # If we are not using axes, then the model_data can be straight plotted...
        if axes is None:
            plt.scatter(x_model_data, y_model_data, label=label, marker=shape)
            if not disable_lines:
                (line,) = plt.plot(
                    x_model_data, y_model_data, label=label, marker=shape
                )
            if trendline:
                (line,) = plt.plot(trendline_x_data, trendline_y_data, label=label)

        #  otherwise, the model_data needs to be plotted on just on axis.
        else:
            axes.scatter(x_model_data, y_model_data, label=label, marker=shape)
            if not disable_lines:
                if colour is None:
                    (line,) = axes.plot(
                        x_model_data, y_model_data, label=label, marker=shape
                    )
                else:
                    (line,) = axes.plot(
                        x_model_data,
                        y_model_data,
                        label=label,
                        marker=shape,
                        color=colour,
                    )
            if trendline:
                if colour is None:
                    (line,) = plt.plot(trendline_x_data, trendline_y_data, label=label)
                else:
                    (line,) = plt.plot(
                        trendline_x_data, trendline_y_data, label=label, color=colour
                    )

    # Set the labels for the axes.
    plt.xlabel(x_label)
    plt.ylabel(y_label)

    return line if not disable_lines else None


def save_figure(figure_name: str, transparent: bool = False) -> None:
    """
    Saves the figure, shuffling existing files out of the way.

    :param figure_name:
        The name of the figure to save.

    :param transparent:
        Whether to save the figure with a transparent background (True) or a white
        background (False).

    """

    # Create a regex for cycling through the files.
    file_regex = re.compile("figure_{}_(?P<old_index>[0-9]).jpg".format(figure_name))

    # Create a storage directory if it doesn't already exist.
    if not os.path.isdir(OLD_FIGURES_DIRECTORY):
        os.mkdir(OLD_FIGURES_DIRECTORY)
    if not os.path.isdir(NEW_FIGURES_DIRECTORY):
        os.mkdir(NEW_FIGURES_DIRECTORY)

    # We need to work download from large numbers to new numbers.
    filenames = sorted(os.listdir(OLD_FIGURES_DIRECTORY))
    filenames.reverse()

    # Incriment all files in the old_figures directory.
    for filename in tqdm(filenames, desc="renaming old files", uint="file"):
        file_match = re.match(file_regex, filename)
        if file_match is None:
            continue
        new_file_name = re.sub(
            str(file_match.group("old_index")),
            str(int(file_match.group("old_index")) + 1),
            filename,
        )
        os.rename(
            os.path.join(OLD_FIGURES_DIRECTORY, filename),
            os.path.join(OLD_FIGURES_DIRECTORY, new_file_name),
        )

    # Move the current _1 file into the old directory
    if os.path.isfile(
        os.path.join(NEW_FIGURES_DIRECTORY, f"figure_{figure_name}_1.jpg")
    ):
        os.rename(
            os.path.join(NEW_FIGURES_DIRECTORY, f"figure_{figure_name}_1.jpg"),
            os.path.join(OLD_FIGURES_DIRECTORY, f"figure_{figure_name}_1.jpg"),
        )

    # If it exists, move the current figure to _1
    if os.path.isfile(os.path.join(NEW_FIGURES_DIRECTORY, f"figure_{figure_name}.jpg")):
        os.rename(
            os.path.join(NEW_FIGURES_DIRECTORY, f"figure_{figure_name}.jpg"),
            os.path.join(NEW_FIGURES_DIRECTORY, f"figure_{figure_name}_1.jpg"),
        )

    # Save the figure
    plt.savefig(
        os.path.join(NEW_FIGURES_DIRECTORY, f"figure_{figure_name}.png"),
        transparent=transparent,
    )


def plot_figure(  # pylint: disable=too-many-branches
    figure_name: str,
    model_data: Dict[Any, Any],
    first_axis_things_to_plot: List[str],
    first_axis_label: str,
    *,
    first_axis_shape: str = "x",
    first_axis_y_limits: Optional[Tuple[int, int]] = None,
    use_data_keys_as_x_axis: bool = False,
    x_axis_thing_to_plot: str = TIME_KEY,
    x_axis_label: str = "Time of day",
    second_axis_things_to_plot: Optional[List[str]] = None,
    second_axis_label: Optional[str] = None,
    second_axis_y_limits: Optional[Tuple[int, int]] = None,
    annotate_maximum: bool = False,
    bar_plot: bool = False,
    disable_lines: bool = False,
    override_axis: Optional[Axes] = None,
    plot_title: Optional[str] = None,
    plot_trendline: bool = False,
    transparent: bool = False,
) -> None:
    """
    Does all the work needed to plot a figure with up to two axes and save it.

    :param figure_name:
        The name to assign to the figure when saving it.

    :param model_data:
        The model_data, as extracted from the JSON outputted by the simulation.

    :param first_axis_things_to_plot:
        The list of variable names (keys in the JSON model_data) to plot on the first axis.

    :param first_axis_label:
        The label to assign to the first y-axis.

    :param first_axis_y_limits:
        A `tuple` giving the lower and upper limits to set for the y axis for the first
        axis.

    :param first_axis_shape:
        A `str` giving an optional override shape for the first axis.

    :param use_data_keys_as_x_axis:
        If set to `True`, then the keys of the data set will be used for the plotting
        rather than the data set specified via the x-axis label.

    :param x_axis_thing_to_plot:
        The variable name to plot on the x-axis.

    :param x_axis_label:
        The label to assign to the x-axis.

    :param x_axis_limits:
        A `tuple` giving the lower and upper limits to set for the x axis.

    :param second_axis_things_to_plot:
        The list of variable names (keys in the JSON model_data) to plot on the second axis.

    :param second_axis_label:
        The label to assign to the second y-axis.

    :param second_axis_y_limits:
        A `tuple` giving the lower and upper limits to set for the y axis for the second
        axis.

    :param annotate_maximum:
        If specified, the maximum will be plotted on the graph.

    :param bar_plot:
        If specified, a bar plot will be generated, rather than a line plot.

    :param disable_lines:
        If specified, lines will be disabled from the plotting.

    :param override_axis:
        If specified, this overrides the fetching of internal axes.

    :param plot_title:
        If specified, a title is addded to the plot.

    :param plot_trendline:
        If `True`, then a best-fit trendline will be added.

    :param transparent:
        Whether to save the figure with a transparent background (True) or with a white
        background (False).

    """

    if override_axis is None:
        _, ax1 = plt.subplots()
    else:
        ax1 = override_axis

    lines = [
        plot(
            entry,
            use_data_keys_as_x_axis,
            x_axis_thing_to_plot,
            x_axis_label,
            first_axis_label,
            model_data,
            hold=True,
            axes=ax1,
            shape=first_axis_shape,
            bar_plot=bar_plot,
            disable_lines=disable_lines,
            trendline=plot_trendline,
        )
        for entry in first_axis_things_to_plot
    ]

    # Adjust the x-tick separation if appropriate.
    locs, _ = plt.xticks()
    if len(locs) > 2 * X_TICK_SEPARATION:
        ax1.set_xticks(ax1.get_xticks()[::X_TICK_SEPARATION])

    if first_axis_y_limits is not None:
        plt.ylim(*first_axis_y_limits)

    # Plot a maximum value box if requested
    if annotate_maximum:
        _annotate_maximum(
            model_data,
            first_axis_things_to_plot,
            ax1,
        )

    # Plot a title if specified.
    if plot_title is not None:
        plt.title(plot_title)

    # Save the figure and return if only one axis is plotted.
    if second_axis_things_to_plot is None:
        ax1.legend(first_axis_things_to_plot)  # , loc="upper left")
        save_figure(figure_name, transparent)
        return

    # Second-axis plotting.
    ax2 = ax1.twinx()

    lines.extend(
        [
            plot(
                entry,
                use_data_keys_as_x_axis,
                x_axis_thing_to_plot,
                x_axis_label,
                second_axis_label,
                model_data,
                hold=True,
                axes=ax2,
                shape=".",
                bar_plot=bar_plot,
                disable_lines=disable_lines,
                trendline=plot_trendline,
            )
            for entry in second_axis_things_to_plot
        ]
    )

    if lines is not None:
        plt.legend(lines, first_axis_things_to_plot + second_axis_things_to_plot)
    else:
        ax2.legend(first_axis_things_to_plot + second_axis_things_to_plot)

    ax2.set_xticks(ax2.get_xticks()[::X_TICK_SEPARATION])

    # Set the y limits if appropriate.
    if second_axis_y_limits is not None:
        plt.ylim(*second_axis_y_limits)

    # Plot a maximum value box if requested
    if annotate_maximum:
        _annotate_maximum(
            model_data,
            second_axis_things_to_plot,
            ax2,
        )

    save_figure(figure_name, transparent)


def plot_two_dimensional_figure(
    figure_name: str,
    logger: Logger,
    model_data: Dict[Any, Any],
    thing_to_plot: str,
    *,
    axis_label: str,
    plot_title: str,
    entry_number: Optional[int] = None,
    hold: bool = False,
    hour: Optional[int] = None,
    minute: Optional[int] = None,
    x_axis_label: str = "Y element index",
) -> None:
    """
    Plots a two-dimensional figure.

    :param figure_name:
        The name to use when saving the figure.

    :param loger:
        The logger used for the analysis module.

    :param model_data:
        The data outputted from the model.

    :param thing_to_plot:
        The name of the variable to plot.

    :param axis_label:
        The label for the y-axis of the plot.

    :param plot_title:
        The title to use for the plot.

    :param entry_number:
        If provided, this is used to compute the entry to plot. Otherwise, hour and
        minute are used.

    :param hold:
        Whether to hold the plot.

    :param hour:
        The hour at which to plot the two-dimensional temperature profile.

    :param minute:
        The minute at which to plot the two-dimensional temperature profile.

    :param x_axis_label:
        The label to use for the x-axis if the plot is 1D.

    """

    # Determine the data index number based on the time.
    if entry_number is None:
        entry_number = int(len(model_data) * ((hour / 24) + (minute / (24 * 60))))
    try:
        data_entry = model_data[entry_number]
    except KeyError:
        try:
            data_entry = model_data[str(entry_number)]
        except KeyError as e:
            logger.error(
                "Key lookup failed for data entry number %s: %s", entry_number, str(e)
            )
            raise

    if thing_to_plot not in data_entry:
        logger.error(
            "Unable to plot %s. Either data is not 2D or name mismatch occured in the "
            "analysis module. See /logs for details.",
            thing_to_plot,
        )
        return

    coordinate_regex = re.compile(r"\((?P<x_index>[0-9]*), (?P<y_index>[0-9]*)\)")

    try:
        x_series = [
            int(re.match(coordinate_regex, coordinate).group("x_index"))
            for coordinate in data_entry[thing_to_plot]
        ]
        y_series = [
            int(re.match(coordinate_regex, coordinate).group("y_index"))
            for coordinate in data_entry[thing_to_plot]
        ]
        z_series = list(data_entry[thing_to_plot].values())
    except AttributeError as e:
        logger.info(str(e))
        logger.error(
            "Unable to match coordinate output for %s. Check output file contents. See "
            "/logs for details.",
            thing_to_plot,
        )
        return

    # Reshape the data for plotting.
    array_shape = (len(set(y_series)), len(set(x_series)))

    # If the data is only 1D, then plot a standard 1D profile through the absorber.
    if 1 in array_shape:
        # If we are not holding the graph, then clear the model_data.
        if not hold:
            plt.clf()
        plt.scatter(y_series, z_series)
        lines = plt.plot(y_series, z_series)
        # Set the labels for the axes.
        plt.xlabel(x_axis_label)
        plt.ylabel(axis_label)
        # Add the legend.
        axis = plt.gca()
        if lines is not None:
            plt.legend(lines, [thing_to_plot])
        else:
            axis.legend([thing_to_plot])
        save_figure(figure_name, False)

        return

    z_array = numpy.zeros(array_shape)
    for index, value in enumerate(z_series):
        z_array[len(set(y_series)) - (y_series[index] + 1), x_series[index]] = value

    # Plot the figure.
    fig3D = plt.figure()
    surface = plt.imshow(
        # axes3D,
        z_array,
        cmap=cm.coolwarm,  # pylint: disable=no-member
        # linewidth=0,
        # antialiased=False,
        aspect=array_shape[1] / array_shape[0],
    )
    plt.title(plot_title)
    plt.xlabel("Element x index")
    plt.ylabel("Element y index")

    # Add axes and colour scale.
    fig3D.colorbar(surface, shrink=0.5, aspect=5, label=axis_label)

    save_figure(figure_name, False)

    x_array = numpy.reshape(x_series, array_shape)
    y_array = numpy.reshape(y_series, array_shape)
    z_array = numpy.reshape(z_series, array_shape)

    # Plot the figure.
    fig3D = plt.figure()
    axes3D = fig3D.gca(projection="3d")
    surface = plt3D.scatter(
        axes3D,
        x_array,
        y_array,
        z_array,
        cmap=cm.coolwarm,  # pylint: disable=no-member
        linewidth=0,
        antialiased=False,
    )

    # Add axes and colour scale.
    fig3D.colorbar(surface, shrink=0.5, aspect=5)
    axes3D.set_xlabel("X index")
    axes3D.set_ylabel("Y index")
    axes3D.set_zlabel(axis_label)

    save_figure(f"{figure_name}_3d", False)
    if not hold:
        plt.close("all")
